from __future__ import annotations

from scitbx.array_family import flex

from dxtbx.model import Beam, Detector, Goniometer, Scan


def test_beam():
    # Construct the object
    b1 = Beam(
        direction=(1, 2, 3),
        wavelength=1.1,
        divergence=0.01,
        sigma_divergence=0.01,
        polarization_normal=(4, 5, 6),
        polarization_fraction=0.9,
        flux=1.0,
        transmission=1.0,
    )

    # Create a dictionary and get the beam back
    d = b1.to_dict()
    b2 = Beam.from_dict(d)
    assert b2 == b1


def test_goniometer():
    # Construct the object
    g1 = Goniometer(
        rotation_axis=(1, 2, 3), fixed_rotation_matrix=(1, 2, 3, 4, 5, 6, 7, 8, 9)
    )

    # Create a dictionary and get the object back
    d = g1.to_dict()
    g2 = Goniometer.from_dict(d)
    assert g2 == g1


def test_scan():
    s1 = Scan(
        image_range=(1, 20),
        oscillation=(5.0, 0.1),
        exposure_times=flex.double(range(20)),
        epochs=flex.double(range(20, 40)),
    )
    s1.set_valid_image_ranges("0", [(1, 20)])

    d = s1.to_dict()
    s2 = Scan.from_dict(d)
    assert s1 == s2


def test_detector():
    d1 = Detector()
    p = d1.add_panel()
    p.set_name("p1")
    p.set_type("panel")
    p.set_pixel_size((0.1, 0.1))
    p.set_image_size((100, 100))
    p.set_trusted_range((0, 1000))
    p.set_local_frame((1, 0, 0), (0, 1, 0), (0, 0, 1))

    p = d1.add_panel()
    p.set_name("p2")
    p.set_type("panel")
    p.set_pixel_size((0.2, 0.2))
    p.set_image_size((200, 200))
    p.set_trusted_range((0, 2000))
    p.set_local_frame((0, 1, 0), (1, 0, 0), (0, 0, 1))

    root = d1.hierarchy()
    g = root.add_group()
    g.set_name("g1")
    g.set_type("group")
    g.set_local_frame((0, 1, 0), (1, 0, 0), (0, 0, 2))
    g.add_panel(d1[0])

    g = root.add_group()
    g.set_name("g2")
    g.set_type("group")
    g.set_local_frame((0, 1, 0), (1, 0, 0), (0, 0, 4))
    g.add_panel(d1[1])

    d = d1.to_dict()
    d2 = Detector.from_dict(d)
    assert len(d1) == len(d2)
    for p1, p2 in zip(d1, d2):
        assert p1 == p2
    assert d1.hierarchy() == d2.hierarchy()
    assert d1 == d2



from __future__ import annotations

import functools
import math

from scitbx import matrix


def _all_equal(a, b):
    return all(x[0] == x[1] for x in zip(a, b))


def _all_approx_equal(a, b, tolerance):
    return all(abs(x[0] - x[1]) < tolerance for x in zip(a, b))


def sequence_diff(sequence1, sequence2, tolerance=None) -> str:
    b_diff = beam_diff
    d_diff = detector_diff
    g_diff = goniometer_diff
    s_diff = scan_diff
    if tolerance:
        b_diff = functools.partial(
            b_diff,
            wavelength_tolerance=tolerance.beam.wavelength,
            direction_tolerance=tolerance.beam.direction,
            polarization_normal_tolerance=tolerance.beam.polarization_normal,
            polarization_fraction_tolerance=tolerance.beam.polarization_fraction,
        )
        d_diff = functools.partial(
            d_diff,
            fast_axis_tolerance=tolerance.detector.fast_axis,
            slow_axis_tolerance=tolerance.detector.slow_axis,
            origin_tolerance=tolerance.detector.origin,
        )
        g_diff = functools.partial(
            g_diff,
            rotation_axis_tolerance=tolerance.goniometer.rotation_axis,
            fixed_rotation_tolerance=tolerance.goniometer.fixed_rotation,
            setting_rotation_tolerance=tolerance.goniometer.setting_rotation,
        )
        s_diff = functools.partial(s_diff, scan_tolerance=tolerance.scan.oscillation)
    output = (
        b_diff(sequence1.get_beam(), sequence2.get_beam()),
        d_diff(sequence1.get_detector(), sequence2.get_detector()),
        g_diff(sequence1.get_goniometer(), sequence2.get_goniometer()),
        s_diff(sequence1.get_scan(), sequence2.get_scan()),
    )

    return "\n".join(block for block in output if block)


def beam_diff(
    beam1,
    beam2,
    wavelength_tolerance=1e-6,
    direction_tolerance=1e-6,
    polarization_normal_tolerance=1e-6,
    polarization_fraction_tolerance=1e-6,
) -> str:
    aw = beam1.get_wavelength()
    bw = beam2.get_wavelength()
    ad = matrix.col(beam1.get_sample_to_source_direction())
    bd = matrix.col(beam2.get_sample_to_source_direction())
    an = matrix.col(beam1.get_polarization_normal())
    bn = matrix.col(beam2.get_polarization_normal())
    af = beam1.get_polarization_fraction()
    bf = beam2.get_polarization_fraction()
    text = []
    if abs(aw - bw) > wavelength_tolerance:
        text.append(f" Wavelength: {aw:f}, {bw:f}")
    if abs(ad.angle(bd)) > direction_tolerance:
        text.append(f" Direction: {tuple(ad)}, {tuple(bd)}")
    if abs(an.angle(bn)) > polarization_normal_tolerance:
        text.append(f" Polarization normal: {tuple(an)}, {tuple(bn)}")
    if abs(af - bf) > polarization_fraction_tolerance:
        text.append(f" Polarization fraction: {af}, {bf}")
    if len(text) > 0:
        text = ["Beam:"] + text
    return "\n".join(text)


def detector_diff(
    detector1,
    detector2,
    fast_axis_tolerance=1e-6,
    slow_axis_tolerance=1e-6,
    origin_tolerance=1e-6,
) -> str:
    text = []
    if len(detector1) != len(detector2):
        text.append("Num panels: %d, %d" % (len(detector1), len(detector2)))
    for i, (aa, bb) in enumerate(zip(detector1, detector2)):
        a_image_size = aa.get_image_size()
        b_image_size = bb.get_image_size()
        a_pixel_size = aa.get_pixel_size()
        b_pixel_size = bb.get_pixel_size()
        a_trusted_range = aa.get_trusted_range()
        b_trusted_range = bb.get_trusted_range()
        a_fast = aa.get_fast_axis()
        b_fast = bb.get_fast_axis()
        a_slow = aa.get_slow_axis()
        b_slow = bb.get_slow_axis()
        a_origin = aa.get_origin()
        b_origin = bb.get_origin()
        temp_text = []
        if not _all_equal(a_image_size, b_image_size):
            temp_text.append(f"  Image size: {a_image_size}, {b_image_size}")
        if not _all_approx_equal(a_pixel_size, b_pixel_size, 1e-7):
            temp_text.append(f"  Pixel size: {a_pixel_size}, {b_pixel_size}")
        if not _all_approx_equal(a_trusted_range, b_trusted_range, 1e-7):
            temp_text.append(f"  Trusted range: {a_trusted_range}, {b_trusted_range}")
        if not _all_approx_equal(a_fast, b_fast, fast_axis_tolerance):
            temp_text.append(f"  Fast axis: {a_fast}, {b_fast}")
        if not _all_approx_equal(a_slow, b_slow, slow_axis_tolerance):
            temp_text.append(f"  Slow axis: {a_slow}, {b_slow}")
        if not _all_approx_equal(a_origin, b_origin, origin_tolerance):
            temp_text.append(f"  Origin: {a_origin}, {b_origin}")
        if len(temp_text) > 0:
            text.append(" panel %d:" % i)
            text.extend(temp_text)
    if len(text) > 0:
        text = ["Detector:"] + text
    return "\n".join(text)


def goniometer_diff(
    goniometer1,
    goniometer2,
    rotation_axis_tolerance=1e-6,
    fixed_rotation_tolerance=1e-6,
    setting_rotation_tolerance=1e-6,
) -> str:
    a_axis = matrix.col(goniometer1.get_rotation_axis())
    b_axis = matrix.col(goniometer2.get_rotation_axis())
    a_fixed = goniometer1.get_fixed_rotation()
    b_fixed = goniometer2.get_fixed_rotation()
    a_setting = goniometer1.get_setting_rotation()
    b_setting = goniometer2.get_setting_rotation()
    text = []
    if abs(a_axis.angle(b_axis)) > rotation_axis_tolerance:
        text.append(f" Rotation axis: {tuple(a_axis)}, {tuple(b_axis)}")
    if not _all_approx_equal(a_fixed, b_fixed, fixed_rotation_tolerance):
        text.append(f" Fixed rotation: {a_fixed}, {b_fixed}")
    if not _all_approx_equal(a_setting, b_setting, setting_rotation_tolerance):
        text.append(f" Setting rotation: {a_setting}, {b_setting}")
    if len(text) > 0:
        text = ["Goniometer:"] + text
    return "\n".join(text)


def scan_diff(scan1, scan2, scan_tolerance=0.03) -> str:
    eps = scan_tolerance * abs(scan1.get_oscillation()[1])
    a_image_range = scan1.get_image_range()
    b_image_range = scan2.get_image_range()
    a_oscillation = scan1.get_oscillation()
    b_oscillation = scan2.get_oscillation()
    a_osc_range = scan1.get_oscillation_range()
    b_osc_range = scan2.get_oscillation_range()

    def mod_2pi(x):
        return x - 2 * math.pi * math.floor(x / (2 * math.pi))

    diff_2pi = abs(mod_2pi(a_osc_range[1]) - mod_2pi(b_osc_range[0]))
    diff_abs = abs(a_osc_range[1] - b_osc_range[0])
    text = []
    if not (a_image_range[1] + 1 == b_image_range[0]):
        text.append(f" Incompatible image range: {a_image_range}, {b_image_range}")
    if abs(a_oscillation[1] - b_oscillation[1]) > eps:
        text.append(f" Incompatible oscillation: {a_oscillation}, {b_oscillation}")
    if min(diff_2pi, diff_abs) > eps * scan1.get_num_images():
        text.append(f" Incompatible oscillation range: {a_osc_range}, {b_osc_range}")
    if len(text) > 0:
        text = ["Scan:"] + text
    return "\n".join(text)



from __future__ import annotations

import pytest

from dxtbx.format.FormatCBFMiniPilatus import FormatCBFMiniPilatus
from dxtbx.model.experiment_list import ExperimentListFactory


def test_single_panel(dials_data, tmpdir):
    filename = dials_data("x4wide") / "X4_wide_M1S4_2_0001.cbf"

    assert FormatCBFMiniPilatus.understand(filename)
    expts = ExperimentListFactory.from_filenames([filename])
    assert len(expts) == 1
    assert len(expts[0].detector) == 1
    assert len(expts[0].imageset.get_raw_data(0)) == 1


def test_multi_panel(dials_data, tmpdir):
    filename = dials_data("x4wide") / "X4_wide_M1S4_2_0001.cbf"

    assert FormatCBFMiniPilatus.understand(filename)
    expts = ExperimentListFactory.from_filenames(
        [filename], format_kwargs={"multi_panel": True}
    )
    assert len(expts) == 1
    assert len(expts[0].detector) == 60
    assert len(expts[0].imageset.get_raw_data(0)) == 60
    expected_xoffset = [-212.47848, -127.51048, -42.54248, 42.42552, 127.39352]
    expected_yoffset = [
        220.00176,
        183.53776,
        147.07376,
        110.60976,
        74.14576,
        37.68176,
        1.21776,
        -35.24624,
        -71.71024,
        -108.17424,
        -144.63824,
        -181.10224,
    ]
    for i, p in enumerate(expts[0].detector):
        assert p.get_image_size() == (487, 195)
        origin = p.get_origin()
        assert origin[0] == pytest.approx(expected_xoffset[i % 5])
        assert origin[1] == pytest.approx(expected_yoffset[i // 5])
        assert origin[2] == pytest.approx(-190.18)



from __future__ import annotations

from collections.abc import Iterator, Sequence
from enum import Enum
from typing import (
    Any,
    TypeVar,
    Union,
    overload,
)

from cctbx.sgtbx import space_group
from cctbx.uctbx import unit_cell
from scitbx.array_family import flex
from scitbx.array_family import shared as flex_shared

# Attempt to use the stub typing for flex-inheritance
from scitbx.array_family.flex import FlexPlain

from dxtbx_model_ext import (
    ExperimentType,
    Probe,  # type: ignore
)

# TypeVar for the set of Experiment models that can be joint-accepted
# - profile, imageset and scalingmodel are handled as 'object'
TExperimentModel = TypeVar(
    "TExperimentModel", BeamBase, Detector, Goniometer, Scan, CrystalBase, object
)

Vec2Float = tuple[float, float]
Vec3Float = tuple[float, float, float]
Vec6Float = tuple[float, float, float, float, float, float]
Vec9Float = tuple[float, float, float, float, float, float, float, float, float]
Vec2Int = tuple[int, int]
Vec4Int = tuple[int, int, int, int]
ScanPropertyTypes = Union[
    flex.int,
    flex.double,
    flex.bool,
    flex.std_string,
    flex.vec2_double,
    flex.vec3_double,
]

class BeamBase:
    @property
    def num_scan_points(self) -> int: ...
    def get_divergence(self) -> float: ...
    def set_divergence(self, divergence: float) -> None: ...
    def get_flux(self) -> float: ...
    def set_flux(self, flux: float) -> None: ...
    def get_num_scan_points(self) -> int: ...
    def get_polarization_fraction(self) -> float: ...
    def set_polarization_fraction(self, polarization_fraction: float) -> None: ...
    def get_polarization_normal(self) -> Vec3Float: ...
    def set_polarization_normal(self, polarization_normal: Vec3Float) -> None: ...
    def get_s0(self) -> Vec3Float: ...
    def set_s0(self, s0: Vec3Float) -> None: ...
    def get_s0_at_scan_point(self, index: int) -> Vec3Float: ...
    def get_s0_at_scan_points(self) -> flex.vec3_double: ...
    def set_s0_at_scan_points(
        self, points: tuple[Vec3Float] | list[Vec3Float]
    ) -> None: ...
    def get_sample_to_source_direction(self) -> Vec3Float: ...
    def get_sample_to_source_distance(self) -> float: ...
    def set_sample_to_source_distance(
        self, sample_to_source_distance: float
    ) -> None: ...
    def get_sigma_divergence(self) -> float: ...
    def set_sigma_divergence(self, sigma_divergence: float) -> None: ...
    def get_transmission(self) -> float: ...
    def set_transmission(self, transmission: float) -> None: ...
    def get_unit_s0(self) -> Vec3Float: ...
    def set_unit_s0(self, unit_s0: Vec3Float) -> None: ...
    def get_wavelength(self) -> float: ...
    def set_wavelength(self, wavelength: float) -> None: ...
    def is_similar_to(
        self,
        other: BeamBase,
        wavelength_tolerance: float = ...,
        direction_tolerance: float = ...,
        polarization_normal_tolerance: float = ...,
        polarization_fraction_tolerance: float = ...,
    ) -> bool: ...
    def reset_scan_points(self) -> None: ...
    def rotate_around_origin(
        self, axis: Vec3Float, angle: float, deg: bool = ...
    ) -> Any: ...
    def set_direction(self, direction: Vec3Float) -> None: ...

class Beam(BeamBase):
    @overload
    def __init__(self, beam: Beam) -> None: ...
    @overload
    def __init__(self, direction: Vec3Float, wavelength: float) -> None: ...
    @overload
    def __init__(self, s0: Vec3Float) -> None: ...
    @overload
    def __init__(
        self,
        direction: Vec3Float,
        wavelength: float,
        divergence: float,
        sigma_divergence: float,
        deg: bool = ...,
    ) -> None: ...
    @overload
    def __init__(
        self, s0: Vec3Float, divergence: float, sigma_divergence: float, deg: bool = ...
    ) -> None: ...
    @overload
    def __init__(
        self,
        direction: Vec3Float,
        wavelength: float,
        divergence: float,
        sigma_divergence: float,
        polarization_normal: Vec3Float,
        polarization_fraction: float,
        flux: float,
        transmission: float,
        probe: Probe = ...,
        deg: bool = ...,
    ) -> None: ...
    @overload
    def __init__(
        self,
        direction: Vec3Float,
        wavelength: float,
        divergence: float,
        sigma_divergence: float,
        polarization_normal: Vec3Float,
        polarization_fraction: float,
        flux: float,
        transmission: float,
        probe: Probe = ...,
        sample_to_source_distance: float = ...,
        deg: bool = ...,
    ) -> None: ...
    @staticmethod
    def from_dict(data: dict) -> Beam: ...
    def to_dict(self) -> dict: ...
    @staticmethod
    def get_probe_from_name(name: str) -> Probe: ...

class PolychromaticBeam(Beam):
    @overload
    def __init__(self, beam: PolychromaticBeam) -> None: ...
    @overload
    def __init__(self, direction: Vec3Float) -> None: ...
    @overload
    def __init__(
        self,
        direction: Vec3Float,
        divergence: float,
        sigma_divergence: float,
        deg: bool = ...,
    ) -> None: ...
    @overload
    def __init__(
        self,
        direction: Vec3Float,
        divergence: float,
        sigma_divergence: float,
        polarization_normal: Vec3Float,
        polarization_fraction: float,
        flux: float,
        transmission: float,
        probe: Probe = ...,
        sample_to_source_distance: float = ...,
        deg: bool = ...,
    ) -> None: ...
    def set_wavelength_range(self, wavelength_range: Vec2Float) -> None: ...
    def get_wavelength_range(self) -> Vec2Float: ...
    @staticmethod
    def from_dict(data: dict) -> PolychromaticBeam: ...
    def to_dict(self) -> dict: ...

class CrystalBase:
    @property
    def num_scan_points(self) -> int: ...
    def change_basis(self, cctbx) -> Any: ...
    def get_A(self) -> Vec9Float: ...
    def get_B(self) -> Vec9Float: ...
    def get_U(self) -> Vec9Float: ...
    def set_A(self, A: Vec9Float) -> None: ...
    def set_B(self, B: Vec9Float) -> None: ...
    def set_U(self, U: Vec9Float) -> None: ...
    def get_A_at_scan_point(self, index: int) -> Vec9Float: ...
    def get_B_at_scan_point(self, index: int) -> Vec9Float: ...
    def get_B_covariance(self) -> flex.double: ...
    def get_B_covariance_at_scan_point(self, index: int) -> flex.double: ...
    def get_B_covariance_at_scan_points(self) -> flex.double: ...
    def get_U_at_scan_point(self, index: int) -> Vec9Float: ...
    def get_cell_parameter_sd(self) -> Vec6Float: ...
    def get_cell_parameter_sd_at_scan_point(self, index: int) -> Vec6Float: ...
    def get_cell_volume_sd(self) -> float: ...
    def get_num_scan_points(self) -> int: ...
    def get_real_space_vectors(self) -> flex.vec3_double: ...
    def get_recalculated_cell_parameter_sd(self) -> Vec6Float: ...
    def get_recalculated_cell_volume_sd(self) -> float: ...
    def get_recalculated_unit_cell(self) -> unit_cell | None: ...
    def get_unit_cell(self) -> unit_cell: ...
    def get_space_group(self) -> space_group: ...
    def get_unit_cell_at_scan_point(self, index: int) -> unit_cell: ...
    def is_similar_to(
        self,
        other: CrystalBase,
        angle_tolerance: float = ...,
        uc_rel_length_tolerance: float = ...,
        uc_abs_angle_tolerance: float = ...,
    ) -> bool: ...
    def reset_scan_points(self) -> None: ...
    def reset_unit_cell_errors(self) -> None: ...
    def rotate_around_origin(
        self, axis: Vec3Float, angle: float, deg: bool = ...
    ) -> Any: ...
    def set_A_at_scan_points(
        self, value: list[Vec9Float] | tuple[Vec9Float] | flex.mat3_double
    ) -> None: ...
    def set_B_covariance(self, covariance: tuple[float, ...] | flex.double) -> None: ...
    def set_B_covariance_at_scan_points(self, data: flex.double) -> None: ...
    def set_recalculated_cell_parameter_sd(self, value: Vec6Float) -> None: ...
    def set_recalculated_cell_volume_sd(self, volume: float) -> None: ...
    def set_recalculated_unit_cell(self, unit_cell: unit_cell) -> None: ...
    def set_space_group(self, space_group: space_group) -> None: ...
    def set_unit_cell(self, unit_cell: unit_cell) -> None: ...
    def update(self, other: CrystalBase) -> None: ...
    def update_B(self) -> None: ...

class Crystal(CrystalBase):
    @overload
    def __init__(self, other: Crystal) -> None: ...
    @overload
    def __init__(
        self,
        real_space_a: float,
        real_space_b: float,
        real_space_c: float,
        space_group: space_group,
    ) -> None: ...
    @overload
    def __init__(
        self,
        real_space_a: float,
        real_space_b: float,
        real_space_c: float,
        space_group_symbol: str,
    ) -> None: ...
    @overload
    def __init__(
        self, A: Vec9Float, space_group: space_group, reciprocal: bool = ...
    ) -> None: ...
    @overload
    def __init__(
        self, A: Vec9Float, space_group_symbol: str, reciprocal: bool = ...
    ) -> None: ...

class MosaicCrystalKabsch2010(Crystal):
    def is_similar_to(
        self,
        other: CrystalBase,
        angle_tolerance: float = ...,
        uc_rel_length_tolerance: float = ...,
        uc_abs_angle_tolerance: float = ...,
        mosaicity_tolerance: float = ...,
    ) -> bool: ...
    def get_mosaicity(self, deg: bool = True) -> float: ...
    def set_mosaicity(self, mosaicity: float, deg: bool = True) -> Any: ...

class MosaicCrystalSauter2014(Crystal):
    def is_similar_to(
        self,
        other: CrystalBase,
        angle_tolerance: float = ...,
        uc_rel_length_tolerance: float = ...,
        uc_abs_angle_tolerance: float = ...,
        half_mosaicity_tolerance: float = ...,
        domain_size_tolerance: float = ...,
    ) -> bool: ...
    def get_half_mosaicity_deg(self) -> float: ...
    def set_half_mosaicity_deg(self, half_mosaicity_deg: float) -> None: ...
    def get_domain_size_ang(self) -> float: ...
    def set_domain_size_ang(self, domain_size_ang: float) -> None: ...

class Detector:
    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, panel: Panel) -> None: ...
    def hierarchy(self) -> DetectorNode: ...
    @overload
    def add_group(self) -> DetectorNode: ...
    @overload
    def add_group(self, group: Panel) -> DetectorNode: ...
    @overload
    def add_panel(self) -> Panel: ...
    @overload
    def add_panel(self, panel: Panel) -> Panel: ...
    @staticmethod
    def from_dict(data: dict) -> Detector: ...
    def to_dict(self) -> dict: ...
    def get_max_inscribed_resolution(self, s0: Vec3Float) -> float: ...
    def get_max_resolution(self, s0: Vec3Float) -> float: ...
    def get_names(self) -> flex.std_string: ...
    def get_panel_intersection(self, s1: Vec3Float) -> int: ...
    def get_ray_intersection(self, s1: Vec3Float) -> tuple[int, Vec2Float]: ...
    def has_projection_2d(self) -> bool: ...
    def is_similar_to(
        self,
        other: Detector,
        fast_axis_tolerance: float = ...,
        slow_axis_tolerance: float = ...,
        origin_tolerance: float = ...,
        static_only: bool = ...,
        ignore_trusted_range: bool = ...,
    ) -> bool: ...
    def rotate_around_origin(
        self, axis: Vec3Float, angle: float, deg: bool = ...
    ) -> Any: ...
    def __getitem__(self, index: int) -> Panel: ...
    def __setitem__(self, index: int, panel: Panel) -> None: ...
    def __len__(self) -> int: ...
    def __iter__(boost) -> Iterator[DetectorNode]: ...

class Experiment:
    beam: BeamBase
    crystal: CrystalBase
    detector: Detector
    goniometer: Goniometer
    identifier: str
    imageset: object
    profile: object
    scaling_model: object
    scan: Scan
    def __init__(
        self,
        beam: BeamBase,
        detector: Detector,
        goniometer: Goniometer,
        scan: Scan,
        crystal: CrystalBase,
        profile: object,
        imageset: object,
        scaling_model: object,
        identifier: str = ...,
    ) -> None: ...
    def is_consistent(self) -> bool: ...
    def is_sequence(self) -> bool: ...
    def is_still(self) -> bool: ...
    def __contains__(self, obj: TExperimentModel) -> bool: ...
    def get_type(self) -> ExperimentType: ...

class ExperimentList:
    @overload
    def __init__(self, experiments: Sequence[Experiment]) -> None: ...
    @overload
    def __init__(self) -> None: ...
    def identifiers(self) -> flex.std_string: ...
    def find(self, identifier: str) -> int: ...
    def append(self, experiment: Experiment) -> None: ...
    def extend(self, other: ExperimentList) -> None: ...
    def clear(self) -> None: ...
    def empty(self) -> bool: ...
    @overload
    def __getitem__(self, index: int) -> Experiment: ...
    @overload
    def __getitem__(self, indices: slice) -> ExperimentList: ...
    def __setitem__(self, index: int, experiment: Experiment) -> None: ...
    def __delitem__(self, index: int) -> None: ...
    def __iter__(self) -> Iterator[Experiment]: ...
    def __contains__(self, obj: TExperimentModel) -> bool: ...
    def replace(self, obj: TExperimentModel, withobj: TExperimentModel) -> None: ...
    def indices(
        self, obj: TExperimentModel | FlexPlain[TExperimentModel]
    ) -> flex.size_t: ...
    def remove_on_experiment_identifiers(self, identifiers: list[str]) -> None: ...
    def select_on_experiment_identifiers(self, identifiers: list[str]) -> None: ...
    def where(
        self,
        beam: BeamBase,
        detector: Detector,
        goniometer: Goniometer,
        scan: Scan,
        profile: object,
        imageset: object,
        scaling_model: object,
    ) -> flex.size_t: ...
    def is_consistent(self) -> bool: ...
    def __len__(self) -> int: ...

class GoniometerBase:
    pass

class Goniometer(GoniometerBase):
    @overload
    def __init__(self, other: Goniometer) -> None: ...
    @overload
    def __init__(self, rotation_axis: Vec3Float) -> None: ...
    @overload
    def __init__(
        self, rotation_axis: Vec3Float, fixed_rotation_matrix: Vec9Float
    ) -> None: ...
    @overload
    def __init__(
        self,
        rotation_axis: Vec3Float,
        fixed_rotation_matrix: Vec9Float,
        setting_rotation_matrix: Vec9Float,
    ) -> None: ...
    @property
    def num_scan_points(self) -> int: ...
    @staticmethod
    def from_dict(data: dict) -> Goniometer: ...
    def get_fixed_rotation(self) -> Vec9Float: ...
    def get_num_scan_points(self) -> int: ...
    def get_rotation_axis(self) -> Vec3Float: ...
    def get_rotation_axis_datum(self) -> Vec3Float: ...
    def get_setting_rotation(self) -> Vec9Float: ...
    def get_setting_rotation_at_scan_point(self, index: int) -> Vec9Float: ...
    def get_setting_rotation_at_scan_points(self) -> flex.mat3_double: ...
    def is_similar_to(
        self,
        other: Goniometer,
        rotation_axis_tolerance: float = ...,
        fixed_rotation_tolerance: float = ...,
        setting_rotation_tolerance: float = ...,
    ) -> bool: ...
    def reset_scan_points(self) -> None: ...
    def rotate_around_origin(
        self, axis: Vec3Float, angle: float, deg: bool = ...
    ) -> None: ...
    def set_fixed_rotation(self, rotation: Vec9Float) -> None: ...
    def set_rotation_axis(self, axis: Vec3Float) -> None: ...
    def set_rotation_axis_datum(self, axis_datum: Vec3Float) -> None: ...
    def set_setting_rotation(self, rotation: Vec9Float) -> None: ...
    def set_setting_rotation_at_scan_points(
        self,
        setting_rotations: list[Vec9Float] | tuple[Vec9Float] | flex.mat3_double,
    ) -> Any: ...
    def to_dict(self) -> dict: ...

class KappaDirection(Enum):
    PlusY = ...
    PlusZ = ...
    MinusY = ...
    MinusZ = ...

class KappaScanAxis(Enum):
    Omega = ...
    Phi = ...

class KappaGoniometer(Goniometer):
    @overload
    def __init__(
        self,
        alpha: float,
        omega: float,
        kappa: float,
        phi: float,
        direction: KappaDirection,
        scan_axis: KappaScanAxis,
    ) -> None: ...
    @overload
    def __init__(
        self,
        alpha: float,
        omega: float,
        kappa: float,
        phi: float,
        direction: str,
        scan_axis: str,
    ) -> None: ...
    def get_alpha_angle(self) -> float: ...
    def get_direction(self) -> KappaDirection: ...
    def get_kappa_angle(self) -> float: ...
    def get_kappa_axis(self) -> Vec3Float: ...
    def get_omega_angle(self) -> float: ...
    def get_omega_axis(self) -> Vec3Float: ...
    def get_phi_angle(self) -> float: ...
    def get_phi_axis(self) -> Vec3Float: ...
    def get_scan_axis(self) -> Vec3Float: ...

class MultiAxisGoniometer(Goniometer):
    def __init__(
        self,
        axes: flex.vec3_double,
        angles: flex.double,
        names: flex.std_string,
        scan_axis: int,
    ) -> None: ...
    @staticmethod
    def from_dict(data: dict) -> MultiAxisGoniometer: ...
    def to_dict(self) -> dict: ...
    def get_angles(self) -> flex.double: ...
    def set_angles(self, angles: flex.double) -> None: ...
    def get_axes(self) -> flex.vec3_double: ...
    def set_axes(self, axes: flex.vec3_double) -> None: ...
    def get_names(self) -> flex.std_string: ...
    def get_scan_axis(self) -> int: ...

class VirtualPanelFrame:
    def get_beam_centre_lab(self, s0: Vec3Float) -> Vec3Float: ...
    def get_beam_centre(self, s0: Vec3Float) -> Vec2Float: ...
    def get_bidirectional_ray_intersection(self, s1: Vec3Float) -> Vec2Float: ...
    def get_d_matrix(self) -> Vec3Float: ...
    def get_D_matrix(self) -> Vec3Float: ...
    def get_directed_distance(self) -> float: ...
    def get_distance(self) -> float: ...
    def get_fast_axis(self) -> Vec3Float: ...
    @overload
    def get_lab_coord(self, xy: Vec2Float) -> Vec3Float: ...
    @overload
    def get_lab_coord(self, xy: flex.vec2_double) -> flex.vec3_double: ...
    def get_local_d_matrix(self) -> Vec9Float: ...
    def get_local_fast_axis(self) -> Vec3Float: ...
    def get_local_origin(self) -> Vec3Float: ...
    def get_local_slow_axis(self) -> Vec3Float: ...
    def get_normal_origin(self) -> Vec2Float: ...
    def get_normal(self) -> Vec3Float: ...
    def get_origin(self) -> Vec3Float: ...
    def get_parent_d_matrix(self) -> Vec9Float: ...
    def get_parent_fast_axis(self) -> Vec3Float: ...
    def get_parent_origin(self) -> Vec3Float: ...
    def get_parent_slow_axis(self) -> Vec3Float: ...
    def get_ray_intersection(self, s1: Vec3Float) -> Vec2Float: ...
    def get_slow_axis(self) -> Vec3Float: ...
    def set_frame(
        self, fast_axis: Vec3Float, slow_axis: Vec3Float, origin: Vec3Float
    ) -> None: ...
    def set_local_frame(
        self, fast_axis: Vec3Float, slow_axis: Vec3Float, origin: Vec3Float
    ) -> None: ...
    def set_parent_frame(
        self, fast_axis: Vec3Float, slow_axis: Vec3Float, origin: Vec3Float
    ) -> None: ...

class VirtualPanel(VirtualPanelFrame):
    def get_name(self) -> str: ...
    def get_type(self) -> str: ...
    def set_name(self, name: str) -> None: ...
    def set_type(self, typename: str) -> None: ...

class PanelData(VirtualPanel):
    def __init__(
        self,
        type: str,
        name: str,
        fast_axis: Vec3Float,
        slow_axis: Vec3Float,
        origin: Vec3Float,
        pixel_size: Vec2Float,
        image_size: Vec2Int,
        trusted_range: Vec2Float,
        thickness: float,
        material: str,
        mu: float = ...,
    ) -> None: ...
    def get_pixel_size(self) -> Vec2Float: ...
    def set_pixel_size(self, size: Vec2Float) -> None: ...
    def get_image_size(self) -> Vec2Int: ...
    def set_image_size(self, size: Vec2Int) -> None: ...
    def get_trusted_range(self) -> Vec2Int: ...
    def set_trusted_range(self, range: Vec2Float) -> None: ...
    def get_thickness(self) -> float: ...
    def set_thickness(self, thickness: float) -> None: ...
    def get_material(self) -> str: ...
    def set_material(self, material: str) -> None: ...
    def get_mu(self) -> float: ...
    def set_mu(self, mu: float) -> None: ...
    def get_raw_image_offset(self) -> Vec2Int: ...
    def set_raw_image_offset(self, offset: Vec2Int) -> None: ...
    def get_mask(self) -> flex_shared.tiny_int_4: ...
    def set_mask(self, mask: flex_shared.tiny_int_4): ...
    def add_mask(self, f0: int, s0: int, f1: int, s1: int) -> None: ...
    def is_similar_to(
        self,
        other: PanelData,
        fast_axis_tolerance: float = ...,
        slow_axis_tolerance: float = ...,
        origin_tolerance: float = ...,
        static_only: bool = ...,
        ignore_trusted_range: bool = ...,
    ) -> bool: ...

class Panel(PanelData):
    @overload
    def __init__(
        self,
        type: str,
        name: str,
        fast_axis: Vec3Float,
        slow_axis: Vec3Float,
        origin: Vec3Float,
        pixel_size: Vec2Float,
        image_size: tuple[int, int],
        trusted_range: Vec2Float,
        thickness: float,
        material: str,
        mu: float = ...,
        identifier: str = ...,
    ) -> None: ...
    @overload
    def __init__(
        self,
        type: str,
        name: str,
        fast_axis: Vec3Float,
        slow_axis: Vec3Float,
        origin: Vec3Float,
        pixel_size: Vec2Float,
        image_size: tuple[int, int],
        trusted_range: Vec2Float,
        thickness: float,
        material: str,
        px_mm: PxMmStrategy,
        mu: float = ...,
        identifier: str = ...,
    ) -> None: ...
    @staticmethod
    @overload
    def from_dict(data: dict) -> Panel: ...
    @staticmethod
    @overload
    def from_dict(data: dict, dx: flex.double, dy: flex.double) -> Panel: ...
    def get_beam_centre_px(self, s0: Vec3Float) -> Vec2Float: ...
    def get_bidirectional_ray_intersection_px(self, s1: Vec3Float) -> Vec2Float: ...
    def get_cos2_two_theta_array(self, s0: Vec3Float) -> flex.double: ...
    def get_gain(self) -> float: ...
    def get_identifier(self) -> str: ...
    def get_image_size_mm(self) -> Vec2Float: ...
    @overload
    def get_max_resolution_at_corners(self, s0: Vec3Float) -> float: ...
    @overload
    def get_max_resolution_at_corners(self, beam: Beam) -> float: ...
    @overload
    def get_max_resolution_at_corners(self, beam: PolychromaticBeam) -> float: ...
    @overload
    def get_max_resolution_ellipse(self, s0: Vec3Float) -> float: ...
    @overload
    def get_max_resolution_ellipse(self, beam: Beam) -> float: ...
    @overload
    def get_max_resolution_ellipse(self, beam: PolychromaticBeam) -> float: ...
    def get_normal_origin_px(self) -> Vec2Float: ...
    def get_pedestal(self) -> float: ...
    def get_pixel_lab_coord(self, px: Vec2Float) -> Vec3Float: ...
    def get_projection_2d(self) -> tuple[Vec4Int, Vec2Int] | tuple[()]: ...
    def get_px_mm_strategy(self) -> PxMmStrategy: ...
    def get_ray_intersection_px(self, s1: Vec3Float) -> Vec2Float: ...
    @overload
    def get_resolution_at_pixel(self, s0: Vec3Float, xy: Vec2Float) -> float: ...
    @overload
    def get_resolution_at_pixel(self, beam: Beam, xy: Vec2Float) -> float: ...
    @overload
    def get_resolution_at_pixel(
        self, beam: PolychromaticBeam, xy: Vec2Float
    ) -> float: ...
    def get_trusted_range_mask(self, image: flex.int | flex.double) -> flex.bool: ...
    def get_two_theta_array(self, s0: Vec3Float) -> flex.double: ...
    def get_two_theta_at_pixel(self, s0: Vec3Float, xy: Vec2Float) -> float: ...
    def get_untrusted_rectangle_mask(self) -> flex.bool: ...
    def is_(self, other: Panel) -> bool: ...
    def is_coord_valid(self, xy: Vec2Float) -> bool: ...
    def is_coord_valid_mm(self, xy: Vec2Float) -> bool: ...
    def is_value_in_trusted_range(self, value: float) -> bool: ...
    @overload
    def millimeter_to_pixel(self, xy: Vec2Float) -> Vec2Float: ...
    @overload
    def millimeter_to_pixel(self, xy: flex.vec2_double) -> flex.vec2_double: ...
    @overload
    def pixel_to_millimeter(self, xy: Vec2Float) -> Vec2Float: ...
    @overload
    def pixel_to_millimeter(self, xy: flex.vec2_double) -> flex.vec2_double: ...
    def rotate_around_origin(
        self, axis: Vec3Float, angle: float, deg: bool = ...
    ) -> None: ...
    def set_gain(self, gain: float) -> None: ...
    def set_identifier(self, identifier: str) -> None: ...
    def set_pedestal(self, pedestal: float) -> None: ...
    def set_projection_2d(self, rotation: Vec4Int, translation: Vec2Int) -> None: ...
    def set_px_mm_strategy(self, strategy: PxMmStrategy) -> None: ...
    def to_dict(self) -> dict: ...

class DetectorNode(Panel):
    @overload
    def add_group(self) -> DetectorNode: ...
    @overload
    def add_group(self, group: Panel) -> DetectorNode: ...
    @overload
    def add_panel(self) -> Panel: ...
    @overload
    def add_panel(self, panel: Panel) -> Panel: ...
    def children(self) -> Iterator[DetectorNode]: ...
    def empty(self) -> bool: ...
    def index(self) -> int: ...
    def is_group(self) -> bool: ...
    def is_panel(self) -> bool: ...
    def is_similar_to(
        self,
        other: PanelData,
        fast_axis_tolerance: float = ...,
        slow_axis_tolerance: float = ...,
        origin_tolerance: float = ...,
        static_only: bool = ...,
        ignore_trusted_range: bool = ...,
    ) -> bool: ...
    def parent(self) -> DetectorNode: ...
    def root(self) -> DetectorNode: ...
    def set_frame(
        self, fast_axis: Vec3Float, slow_axis: Vec3Float, origin: Vec3Float
    ) -> None: ...
    def set_local_frame(
        self, fast_axis: Vec3Float, slow_axis: Vec3Float, origin: Vec3Float
    ) -> None: ...
    def set_parent_frame(
        self, fast_axis: Vec3Float, slow_axis: Vec3Float, origin: Vec3Float
    ) -> None: ...
    def __getitem__(self, index: int) -> DetectorNode: ...
    def __iter__(boost) -> Iterator[DetectorNode]: ...
    def __len__(self) -> int: ...

class PxMmStrategy:
    def name(self) -> str: ...
    def to_millimeter(self, panel: Panel, xy: Vec2Float) -> Vec2Float: ...
    def to_pixel(self, panel: Panel, xy: Vec2Float) -> Vec2Float: ...

class SimplePxMmStrategy(PxMmStrategy):
    pass

class ParallaxCorrectedPxMmStrategy(PxMmStrategy):
    def __init__(self, mu: float, t0: float) -> None: ...
    def mu(self) -> float: ...
    def t0(self) -> float: ...

class OffsetPxMmStrategy(PxMmStrategy):
    def __init__(self, dx: flex.double, dy: flex.double) -> None: ...
    def dx(self) -> flex.double: ...
    def dy(self) -> flex.double: ...

class OffsetParallaxCorrectedPxMmStrategy(ParallaxCorrectedPxMmStrategy):
    def __init__(
        self, mu: float, t0: float, dx: flex.double, dy: flex.double
    ) -> None: ...
    def dx(self) -> flex.double: ...
    def dy(self) -> flex.double: ...

class ScanBase:
    pass

class Scan(ScanBase):
    @overload
    def __init__(
        self,
        image_range: Vec2Int,
        oscillation: Vec2Float,
        batch_offset: int,
        deg: bool = ...,
    ) -> None: ...
    @overload
    def __init__(
        self,
        image_range: Vec2Int,
        oscillation: Vec2Float,
        exposure_times: flex.double,
        epochs: flex.double,
        batch_offset: int,
        deg: bool = ...,
    ) -> None: ...
    @overload
    def __init__(
        self,
        image_range: Vec2Int,
        properties_table: dict,
        batch_offset: int,
        deg: bool = ...,
    ) -> None: ...
    def append(self, other: Scan, scan_tolerance: float) -> None: ...
    @staticmethod
    def from_dict(data: dict) -> Scan: ...
    @overload
    def get_angle_from_array_index(self, index: float, deg: bool = ...) -> float: ...
    @overload
    def get_angle_from_array_index(
        self, index: flex.double, deg: bool = ...
    ) -> flex.double: ...
    def get_angle_from_image_index(self, index: float, deg: bool = ...) -> float: ...
    @overload
    def get_array_index_from_angle(self, angle: float, deg: bool = ...) -> float: ...
    @overload
    def get_array_index_from_angle(
        self, angle: flex.double, deg: bool = ...
    ) -> flex.double: ...
    def get_array_indices_with_angle(
        self, angle: float, deg: bool = ...
    ) -> flex.vec2_double: ...
    def get_array_range(self) -> Vec2Int: ...
    def get_batch_for_array_index(self, index: int) -> int: ...
    def get_batch_for_image_index(self, index: int) -> int: ...
    def get_batch_offset(self) -> int: ...
    def set_batch_offset(self, batch_offset: int) -> int: ...
    def get_batch_range(self) -> Vec2Int: ...
    def get_epochs(self) -> flex.double: ...
    def set_epochs(self, epochs: flex.double) -> None: ...
    def get_exposure_times(self) -> flex.double: ...
    def set_exposure_times(self, times: flex.double) -> None: ...
    def get_image_epoch(self, index: int) -> float: ...
    def get_image_index_from_angle(self, angle: float, deg: bool = ...) -> float: ...
    def get_image_indices_with_angle(
        self, angle: float, deg: bool = ...
    ) -> flex.vec2_double: ...
    def get_image_oscillation(self, index: int, deg: bool = ...) -> Vec2Float: ...
    def get_image_range(self) -> Vec2Int: ...
    def set_image_range(self, image_range: Vec2Int) -> None: ...
    def get_num_images(self) -> int: ...
    def get_oscillation(self, deg: bool = ...) -> Vec2Float: ...
    def set_oscillation(self, oscillation: Vec2Float, deg: bool = ...) -> None: ...
    def get_oscillation_range(self, deg: bool = ...) -> Vec2Float: ...
    def get_valid_image_ranges(self, i: str) -> list[Vec2Int]: ...
    def set_valid_image_ranges(self, i: str, ranges: list[Vec2Int]) -> None: ...
    def get_properties(self) -> dict: ...
    def set_properties(self, properties: dict) -> None: ...
    def has_property(self, key: str) -> bool: ...
    def get_property(self, key: str) -> ScanPropertyTypes: ...
    @overload
    def set_property(self, key: str, value: flex.double) -> None: ...
    @overload
    def set_property(self, key: str, value: flex.bool) -> None: ...
    @overload
    def set_property(self, key: str, value: flex.int) -> None: ...
    @overload
    def set_property(self, key: str, value: flex.std_string) -> None: ...
    @overload
    def set_property(self, key: str, value: flex.vec2_double) -> None: ...
    @overload
    def set_property(self, key: str, value: flex.vec3_double) -> None: ...
    @overload
    def is_angle_valid(self, angle: float, deg: bool = ...) -> bool: ...
    @overload
    def is_angle_valid(self, angle: flex.double, deg: bool = ...) -> flex.bool: ...
    def is_array_index_valid(self, index: int) -> bool: ...
    def is_batch_valid(self, batch: int) -> bool: ...
    def is_image_index_valid(self, index: int) -> bool: ...
    def is_still(self) -> bool: ...
    def swap(self, other: Scan) -> None: ...
    def to_dict(self) -> dict: ...
    def __getitem__(self, index: int | slice) -> Scan: ...
    def __len__(self) -> int: ...
    def __add__(self, other: Scan) -> Scan: ...
    def __iadd__(self, other: Scan) -> Scan: ...
    def __ge__(self, other: Scan) -> bool: ...
    def __gt__(self, other: Scan) -> bool: ...
    def __le__(self, other: Scan) -> bool: ...
    def __lt__(self, other: Scan) -> bool: ...

class Spectrum:
    @overload
    def __init__(self, other: Spectrum) -> None: ...
    @overload
    def __init__(self, energies: flex.double, weights: flex.double) -> None: ...
    @staticmethod
    def from_dict(data: dict) -> Spectrum: ...
    def get_emax_eV(self) -> float: ...
    def get_emin_eV(self) -> float: ...
    def get_energies_eV(self) -> flex.double: ...
    def get_weighted_energy_eV(self) -> float: ...
    def get_weighted_energy_variance(self) -> float: ...
    def get_weighted_wavelength(self) -> float: ...
    def get_weights(self) -> flex.double: ...
    def to_dict(self) -> dict: ...

class flex_Beam(FlexPlain[Beam]):
    pass

class flex_Spectrum(FlexPlain[Spectrum]):
    pass

def get_mod2pi_angles_in_range(
    range: Vec2Float, angle: float, deg: bool = False
) -> flex.double: ...
def is_angle_in_range(range: Vec2Float, angle: float, deg: bool = False) -> bool: ...
def get_range_of_mod2pi_angles(
    range: Vec2Float, angle: float, deg: bool = False
) -> Vec2Float: ...
@overload
def parallax_correction(
    d: float, la: float, xy0: Vec2Float, xy: Vec2Float
) -> Vec2Float: ...
@overload
def parallax_correction(
    mu: float,
    t0: float,
    xy: Vec2Float,
    fast_axis: Vec3Float,
    slow_axis: Vec3Float,
    origin: Vec3Float,
) -> Vec2Float: ...
@overload
def parallax_correction_inv(
    d: float, la: float, xy0: Vec2Float, xy: Vec2Float
) -> Vec2Float: ...
@overload
def parallax_correction_inv(
    mu: float,
    t0: float,
    xy: Vec2Float,
    fast_axis: Vec3Float,
    slow_axis: Vec3Float,
    origin: Vec3Float,
) -> Vec2Float: ...



from __future__ import annotations

import h5py
import pytest

import dxtbx.format.Registry as Registry
from dxtbx.format.Format import Format, abstract
from dxtbx.format.FormatHDF5 import FormatHDF5


@abstract
class AbstractFormatTest(Format):
    schemes = ["test"]

    @classmethod
    def understand(cls):
        return True


class FormatTest(AbstractFormatTest):
    @classmethod
    def understand(cls):
        return True


def test_abstract(tmp_path):
    assert AbstractFormatTest.is_abstract()
    assert not FormatTest.is_abstract()

    # Write an empty h5 file so that it won't match
    with h5py.File(tmp_path / "test.h5", "w"):
        pass
    assert FormatHDF5.is_abstract()
    with pytest.raises(TypeError):
        FormatHDF5(str(tmp_path / "test.h5"))

    assert Registry.get_format_class_for_file(str(tmp_path / "test.h5")).is_abstract()



from __future__ import annotations

from pathlib import Path

from dxtbx.model.detector import DetectorFactory


def test_detector():
    """A test class for the detector class."""

    assert DetectorFactory.simple(
        "CCD",
        100.0,
        (45.0, 52.0),
        "+x",
        "-y",
        (0.172, 0.172),
        (516, 590),
        (0, 1024),
        [],
    )
    assert DetectorFactory.two_theta(
        "CCD",
        60.0,
        (35.0, 34.0),
        "+x",
        "+y",
        "+x",
        30,
        (0.07, 0.07),
        (1042, 1042),
        (0, 1024),
        [],
    )

    image = Path(__file__).parent / "phi_scan_001.cbf"

    assert DetectorFactory.imgCIF(str(image), "CCD")
    # x = DetectorFactory.XDS(xparm)



from __future__ import annotations

from dxtbx.command_line import show_mask_info


def test_show_mask_info(dials_data, capsys):
    data = dials_data("image_examples") / "dectris_eiger_master.h5"

    show_mask_info.run([str(data)])
    captured = capsys.readouterr()
    assert not captured.err
    assert "Module 0 has 637992 masked pixels of 10166590" in captured.out



from __future__ import annotations

import random

from dxtbx.model import (
    get_mod2pi_angles_in_range,
    get_range_of_mod2pi_angles,
    is_angle_in_range,
)


def test_is_angle_in_random_range():
    """Test that for a range of angles and angular ranges, the
    is_angle_in_range function correctly calculates if the angle
    is in the range."""

    # Create a number of ranges between 0 and 360 to see if they are within range
    num_range = 100
    for n in range(num_range):
        angular_range = (int(random.random() * 360), int(random.random()) * 360)

        # If A < B or A > B
        if angular_range[0] < angular_range[1]:
            # Check that the following are true
            #   angle in range 0 -> A = False
            #   angle in range A -> B = True
            #   angle in range B -> 360 = False
            for angle in range(0, angular_range[0]):
                assert is_angle_in_range(angular_range, angle, True) is False
            for angle in range(angular_range[0], angular_range[1] + 1):
                assert is_angle_in_range(angular_range, angle, True) is True
            for angle in range(angular_range[1] + 1, 360):
                assert is_angle_in_range(angular_range, angle, True) is False
        else:
            # Check that the following are true
            #   angle in range 0 -> B = True
            #   angle in range B -> A = False
            #   angle in range A -> 360 = True
            for angle in range(0, angular_range[1] + 1):
                assert is_angle_in_range(angular_range, angle, True) is True
            for angle in range(angular_range[1] + 1, angular_range[0]):
                assert is_angle_in_range(angular_range, angle, True) is False
            for angle in range(angular_range[0], 360):
                assert is_angle_in_range(angular_range, angle, True) is True


def test_is_angle_sequence_in_range():
    # Create a range over 360 and make sure all angles are valid
    angular_range = (-10, 370)
    for angle in range(0, 360):
        assert is_angle_in_range(angular_range, angle, True) is True


def test_get_range_of_mod2pi_angles():
    """Get the range of equivalent within a given angular range."""
    # In a 360 deg range, have only 1 angle
    a0, a1 = get_range_of_mod2pi_angles((0, 360), 180, deg=True)
    assert a0 == 180 and a1 == 180

    # If no angles within range, a0 > a1
    a0, a1 = get_range_of_mod2pi_angles((181, 360), 180, deg=True)
    assert a0 > a1

    # With 720 deg range, have 2 angles
    a0, a1 = get_range_of_mod2pi_angles((0, 720), 180, deg=True)
    assert a0 == 180 and a1 == 540

    # With 1080 deg range, have 3 angles
    a0, a1 = get_range_of_mod2pi_angles((-360, 720), 180, deg=True)
    assert a0 == -180 and a1 == 540


def test_get_mod2pi_angles_in_range():
    """Get the list of equivalent within a given angular range."""
    # In a 360 deg range, have only 1 angle
    a = get_mod2pi_angles_in_range((0, 360), 180, deg=True)
    assert len(a) == 1 and a[0] == 180

    # If no angles within range have no angles
    a = get_mod2pi_angles_in_range((181, 360), 180, deg=True)
    assert len(a) == 0

    # With 720 deg range, have 2 angles
    a = get_mod2pi_angles_in_range((0, 720), 180, deg=True)
    assert len(a) == 2 and a[0] == 180 and a[1] == 540

    # With 1080 deg range, have 3 angles
    a = get_mod2pi_angles_in_range((-360, 720), 180, deg=True)
    assert len(a) == 3 and a[0] == -180 and a[1] == 180 and a[2] == 540



from __future__ import annotations

import h5py

import libtbx

from dxtbx.format.FormatNXmxDLS import FormatNXmxDLS
from dxtbx.masking import GoniometerMaskerFactory
from dxtbx.model import MultiAxisGoniometer

# These are the instrument names that should be used according to
# https://manual.nexusformat.org/classes/applications/NXmx.html and
# https://mmcif.wwpdb.org/dictionaries/mmcif_pdbx_v50.dic/Items/_diffrn_source.type.html
VALID_NAMES = {
    # "long" names
    "DIAMOND BEAMLINE I03",
    "DIAMOND BEAMLINE I04",
    # "short" names
    "DLS I03",
    "DLS I04",
    # "legacy" names used until 2021/03/12
    "I03",
    "I04",
}


class FormatNXmxDLS16M(FormatNXmxDLS):
    @staticmethod
    def understand(image_file):
        # Get the file handle
        with h5py.File(image_file) as handle:
            name = FormatNXmxDLS16M.get_instrument_name(handle)
            return name and name.upper() in VALID_NAMES

    def has_dynamic_shadowing(self, **kwargs):
        dynamic_shadowing = kwargs.get("dynamic_shadowing", False)
        if not isinstance(self.get_goniometer(), MultiAxisGoniometer):
            # Single-axis goniometer, no goniometer shadows
            return False
        if dynamic_shadowing in (libtbx.Auto, "Auto"):
            return True
        return dynamic_shadowing

    def __init__(self, image_file, **kwargs):
        """Initialise the image structure from the given file."""

        super().__init__(image_file, **kwargs)
        self._dynamic_shadowing = self.has_dynamic_shadowing(**kwargs)

    def get_goniometer_shadow_masker(self, goniometer=None):
        if not self._dynamic_shadowing:
            return None

        if goniometer is None:
            goniometer = self.get_goniometer()

        assert goniometer is not None

        if goniometer.get_names()[1] == "chi":
            return GoniometerMaskerFactory.smargon(goniometer)
        elif goniometer.get_names()[1] == "kappa":
            return GoniometerMaskerFactory.mini_kappa(goniometer)

        else:
            raise RuntimeError(
                f"Don't understand this goniometer: {list(goniometer.get_names())}"
            )



from __future__ import annotations

import itertools
import logging
from typing import Literal, Optional

import h5py
import numpy as np
import nxmx

import cctbx
from cctbx import eltbx
from scitbx.array_family import flex

import dxtbx.model
from dxtbx import flumpy

logger = logging.getLogger(__name__)


KNOWN_SENSOR_MATERIALS = {
    "Si": "Si",
    "Silicon": "Si",
    "CdTe": "CdTe",
    "GaAs": "GaAs",
}


# Conversion from the McStas coordinate system as used by NeXus to the imgCIF
# coordinate system conventionally used by dxtbx:
#   https://manual.nexusformat.org/design.html#design-coordinatesystem
#   https://www.iucr.org/__data/iucr/cifdic_html/2/cif_img.dic/Caxis.html
MCSTAS_TO_IMGCIF = np.diag([-1, 1, -1])


def get_dxtbx_goniometer(nxsample: nxmx.NXsample) -> dxtbx.model.Goniometer | None:
    """Generate a dxtbx goniometer model from an NXsample.

    If the NXsample doesn't have a valid depends_on field, then return None.
    """
    if not nxsample.depends_on:
        return None
    dependency_chain = nxmx.get_dependency_chain(nxsample.depends_on)
    logger.debug("Sample dependency chain: %s", dependency_chain)
    axes = nxmx.get_rotation_axes(dependency_chain)
    if len(axes.axes) == 1:
        return dxtbx.model.GoniometerFactory.make_goniometer(
            MCSTAS_TO_IMGCIF @ axes.axes[0], np.identity(3).flatten()
        )
    else:
        if np.sum(axes.is_scan_axis) == 0:
            # A sequence of still images, choose an arbitrary scan axis
            scan_axis = 0
        else:
            assert np.sum(axes.is_scan_axis) == 1, "only one scan axis is supported"
            scan_axis = int(np.where(axes.is_scan_axis)[0][0])
        return dxtbx.model.GoniometerFactory.make_multi_axis_goniometer(
            flex.vec3_double((MCSTAS_TO_IMGCIF @ axes.axes.T).T),
            flex.double(axes.angles),
            flex.std_string(axes.names),
            scan_axis,
        )


class CachedWavelengthBeamFactory:
    """Defer Beam generation whilst caching the wavelength value"""

    def __init__(self, nxbeam: nxmx.NXbeam):
        self.handle = nxbeam._handle
        self.index = None
        self.model = None
        self.spectrum = None

    def make_beam(self, index: int = 0) -> dxtbx.model.Beam:
        self.read_models(index)
        return self.model

    def make_spectrum(self, index: int = 0) -> dxtbx.model.Spectrum:
        self.read_models(index)
        return self.spectrum

    def read_models(self, index: int = 0):
        # Cached model
        if self.model is not None and index == self.index:
            return

        # Get the items from the NXbeam class
        # Note it would be better if there were a general way to read weights
        # and variants from the nxmx classes
        # See https://github.com/cctbx/dxtbx/issues/549
        primary_key = "incident_wavelength"
        wavelength = self.handle[primary_key]
        spectrum_wavelengths = wavelength
        spectrum_weights = self.handle.get(primary_key + "_weights")
        if spectrum_weights is None:
            # Handle deprecation: https://github.com/nexusformat/definitions/issues/837
            spectrum_weights = self.handle.get(primary_key + "_weight")

        # If the wavelength array does not represent spectra, look for spectra
        # in the variant chain
        variant_test = wavelength
        has_variant_spectra = False
        while spectrum_weights is None:
            if "variant" in variant_test.attrs:
                variant_key = variant_test.attrs["variant"]
                variant_wavelengths = self.handle[variant_key]
                variant_weights = self.handle.get(variant_key + "_weights")
                if variant_weights is None:
                    # Handle deprecation: https://github.com/nexusformat/definitions/issues/837
                    variant_weights = self.handle.get(variant_key + "_weight")
                if variant_weights is None:
                    variant_test = variant_wavelengths  # Keep looking
                else:
                    # Found spectra
                    spectrum_wavelengths = variant_wavelengths
                    spectrum_weights = variant_weights  # cause while loop to end
                    has_variant_spectra = True
            else:
                break

        if index is None:
            index = 0
        self.index = index

        def get_wavelength(wavelength):
            if wavelength.shape in ((), (1,)):
                wavelength_value = wavelength[()].item()
            else:
                wavelength_value = wavelength[index]
            wavelength_units = nxmx.units(wavelength)
            wavelength_value = float(
                (wavelength_value * wavelength_units).to("angstrom").magnitude
            )
            return wavelength_value

        if spectrum_weights is None:
            # Construct the beam model
            wavelength_value = get_wavelength(wavelength)
            self.model = dxtbx.model.Beam(
                direction=(0, 0, 1), wavelength=wavelength_value
            )
        else:
            self.model = dxtbx.model.Beam()
            self.model.set_direction((0, 0, 1))

            wavelength_units = nxmx.units(spectrum_wavelengths)

            if len(spectrum_wavelengths.shape) > 1:
                spectrum_wavelengths = spectrum_wavelengths[index]
            else:
                spectrum_wavelengths = spectrum_wavelengths[()]
            if len(spectrum_weights.shape) > 1:
                spectrum_weights = spectrum_weights[index]
            else:
                spectrum_weights = spectrum_weights[()]

            spectrum_wavelengths = (
                (spectrum_wavelengths * wavelength_units).to("angstrom").magnitude
            )
            spectrum_energies = cctbx.factor_ev_angstrom / spectrum_wavelengths
            self.spectrum = dxtbx.model.Spectrum(spectrum_energies, spectrum_weights)

            if has_variant_spectra:
                wavelength_value = get_wavelength(wavelength)
                self.model.set_wavelength(wavelength_value)
            else:
                self.model.set_wavelength(self.spectrum.get_weighted_wavelength())


def get_dxtbx_scan(
    nxsample: nxmx.NXsample, nxdetector: nxmx.NXdetector
) -> dxtbx.model.Scan | None:
    """Generate a dxtbx scan model from an NXsample.

    If the NXsample doesn't have a valid depends_on field, then return None.
    """
    if not nxsample.depends_on:
        return None
    dependency_chain = nxmx.get_dependency_chain(nxsample.depends_on)
    logger.debug("Sample dependency chain: %s", dependency_chain)
    scan_axis = None
    for t in dependency_chain:
        # Find the first varying rotation axis
        if (
            t.transformation_type == "rotation"
            and len(t) > 1
            and not np.all(t[()] == t[0])
        ):
            scan_axis = t
            break

    if scan_axis is None:
        # Fall back on the first varying axis of any type
        for t in dependency_chain:
            if len(t) > 1 and not np.all(t[()] == t[0]):
                scan_axis = t
                break

    if scan_axis is None:
        scan_axis = nxsample.depends_on

    is_rotation = scan_axis.transformation_type == "rotation"
    num_images = len(scan_axis)
    image_range = (1, num_images)

    oscillation = (0, 0)
    if is_rotation:
        start = scan_axis[0].to("degree")
        if scan_axis.end:
            steps = scan_axis.end[()] - scan_axis[()]
        elif num_images > 1:
            steps = np.diff(scan_axis[()])
        else:
            steps = nxmx.ureg.Quantity(0, "degree")

        step = np.median(steps).to("degree")
        if np.any(np.abs(steps - step) > abs(0.1 * step)):
            logger.warning(
                "One or more recorded oscillation widths differ from the median "
                "by more than 10%. The rotation axis of your goniometer may not "
                "have been scanning at a constant speed throughout the data "
                "collection."
            )

        oscillation = tuple(float(f) for f in (start.magnitude, step.magnitude))

    if nxdetector.frame_time is not None:
        frame_time = float(nxdetector.frame_time.to("seconds").magnitude)
        exposure_times = flex.double(num_images, frame_time)
        epochs = flex.double_range(0, num_images) * frame_time
    else:
        exposure_times = flex.double(num_images, 0)
        epochs = flex.double(num_images, 0)

    return dxtbx.model.Scan(
        image_range,
        oscillation,
        exposure_times,
        epochs,
        batch_offset=0,
        deg=True,
    )


def get_dxtbx_detector(
    nxdetector: nxmx.NXdetector,
    wavelength: float,
    nxdata: nxmx.NXdata | None = None,
) -> dxtbx.model.Detector:
    """Generate a dxtbx detector model from an NXdetector and NXbeam.

    If the NXdetector contains multiple NXdetector_modules, then a hierarchical detector
    will be returned, else a "flat" detector model with a single panel will be returned
    where there is only a single NXdetector_module.
    """

    detector = dxtbx.model.Detector()

    root: dxtbx.model.Detector | dxtbx.model.Panel
    if len(nxdetector.modules) > 1:
        root = detector.hierarchy()
    else:
        root = detector

    for module in nxdetector.modules:
        if len(nxdetector.modules) > 1:
            # Set up the detector hierarchy
            if module.fast_pixel_direction.depends_on is not None:
                reversed_dependency_chain = list(
                    reversed(
                        nxmx.get_dependency_chain(
                            module.fast_pixel_direction.depends_on
                        )
                    )
                )
                pg: dxtbx.model.Detector | dxtbx.model.Panel | None = None

                # Verify that equipment_components in the dependency chain are
                # 1) contiguous and 2) unique
                found = []
                for dependency in reversed_dependency_chain:
                    if dependency.equipment_component:
                        if dependency.equipment_component in found:
                            assert dependency.equipment_component == found[-1]
                        found.append(dependency.equipment_component)

                # Group any transformations together that share the same equipment_component
                # to reduce the number of hierarchy levels

                # Keep transformations without equipment_component set separate by using
                # a different key
                counter = 0

                def equipment_component_key(dependency):
                    if dependency.equipment_component:
                        return dependency.equipment_component  # always a string
                    else:
                        nonlocal counter
                        counter += 1
                        return counter

                for _, transformation_group in itertools.groupby(
                    reversed_dependency_chain, key=equipment_component_key
                ):
                    transformation_group = list(transformation_group)
                    name = transformation_group[-1].path
                    if pg is None:
                        pg = root
                    else:
                        assert isinstance(
                            pg, (dxtbx.model.Detector, dxtbx.model.DetectorNode)
                        )
                        pg_names = [child.get_name() for child in pg]
                        if name in pg_names:
                            pg = pg[
                                pg_names.index(name)
                            ]  # Getitem always returns panel
                            continue
                        else:
                            pg = pg.add_group()
                    A = nxmx.get_cumulative_transformation(transformation_group)
                    origin = MCSTAS_TO_IMGCIF @ A[0, :3, 3]
                    fast = (
                        MCSTAS_TO_IMGCIF @ (A @ np.array((-1, 0, 0, 1)))[0, :3] - origin
                    )
                    slow = (
                        MCSTAS_TO_IMGCIF @ (A @ np.array((0, 1, 0, 1)))[0, :3] - origin
                    )
                    pg.set_local_frame(fast, slow, origin)
                    assert name is not None
                    pg.set_name(name)
                # assert pg is not None
        else:
            # Use a flat detector model
            pg = root

        pixel_size = (
            module.fast_pixel_direction[()].to("mm").magnitude.item(),
            module.slow_pixel_direction[()].to("mm").magnitude.item(),
        )

        if isinstance(pg, dxtbx.model.DetectorNode):
            # Hierarchical detector model
            fast_axis = MCSTAS_TO_IMGCIF @ module.fast_pixel_direction.vector
            slow_axis = MCSTAS_TO_IMGCIF @ module.slow_pixel_direction.vector
            origin = MCSTAS_TO_IMGCIF @ (
                module.fast_pixel_direction.offset.to("mm").magnitude
                if module.fast_pixel_direction.offset is not None
                else (
                    np.array([0.0, 0.0, 0.0])
                    + module.slow_pixel_direction.offset.to("mm").magnitude
                    if module.slow_pixel_direction.offset is not None
                    else np.array([0.0, 0.0, 0.0])
                )
            )
        else:
            # Flat detector model

            # Apply any rotation components of the dependency chain to the fast axis
            assert module.fast_pixel_direction.depends_on is not None
            fast_axis_depends_on = [
                t
                for t in nxmx.get_dependency_chain(
                    module.fast_pixel_direction.depends_on
                )
                if t.transformation_type == "rotation"
            ]
            if fast_axis_depends_on:
                R = nxmx.get_cumulative_transformation(fast_axis_depends_on)[0, :3, :3]
            else:
                R = np.identity(3)
            fast_axis = MCSTAS_TO_IMGCIF @ R @ module.fast_pixel_direction.vector

            # Apply any rotation components of the dependency chain to the slow axis
            assert module.slow_pixel_direction.depends_on is not None
            slow_axis_depends_on = [
                t
                for t in nxmx.get_dependency_chain(
                    module.slow_pixel_direction.depends_on
                )
                if t.transformation_type == "rotation"
            ]
            if slow_axis_depends_on:
                R = nxmx.get_cumulative_transformation(slow_axis_depends_on)[0, :3, :3]
            else:
                R = np.identity(3)
            slow_axis = MCSTAS_TO_IMGCIF @ R @ module.slow_pixel_direction.vector

            # Apply all components of the dependency chain to the module offset to get the
            # dxtbx panel origin
            dependency_chain = nxmx.get_dependency_chain(
                module.fast_pixel_direction.depends_on
            )
            A = nxmx.get_cumulative_transformation(dependency_chain)

            origin = MCSTAS_TO_IMGCIF @ (
                (
                    module.fast_pixel_direction.offset.to("mm").magnitude
                    if module.fast_pixel_direction.offset is not None
                    else (
                        np.array([0.0, 0.0, 0.0])
                        + module.slow_pixel_direction.offset.to("mm").magnitude
                        if module.slow_pixel_direction.offset is not None
                        else np.array([0.0, 0.0, 0.0])
                    )
                )
                + A[0, :3, 3]
            )

            if (
                origin[0] == 0
                and origin[1] == 0
                and nxdetector.beam_center_x
                and nxdetector.beam_center_y
            ):
                # fallback on the explicit beam center - this is needed for some older
                # dectris eiger filewriter datasets
                origin -= nxdetector.beam_center_x.magnitude * pixel_size[0] * fast_axis
                origin -= nxdetector.beam_center_y.magnitude * pixel_size[1] * slow_axis

        # dxtbx requires image size in the order fast, slow - which is the reverse of what
        # is stored in module.data_size. Additionally, data_size can have more than 2
        # dimensions, for multi-module detectors. So take the last two dimensions and reverse
        # them.  Examples:
        # [1,2,3]   --> (3, 2)
        # [1,2]     --> (2, 1)
        image_size = (int(module.data_size[-1]), int(module.data_size[-2]))
        underload = (
            float(nxdetector.underload_value)
            if nxdetector.underload_value is not None
            else -0x7FFFFFFF
        )
        overload = (
            float(nxdetector.saturation_value)
            if nxdetector.saturation_value is not None
            else 0x7FFFFFFF
        )
        # The dxtbx trusted_range is inclusive [min-trusted-value, max-trusted-value]
        trusted_range = (underload, overload)

        material = KNOWN_SENSOR_MATERIALS.get(nxdetector.sensor_material)
        if not material:
            raise ValueError(f"Unknown material: {nxdetector.sensor_material}")
        thickness = float(nxdetector.sensor_thickness.to("mm").magnitude)
        table = eltbx.attenuation_coefficient.get_table(material)
        mu = table.mu_at_angstrom(wavelength) / 10.0
        px_mm = dxtbx.model.ParallaxCorrectedPxMmStrategy(mu, thickness)
        name = module.path

        assert name is not None
        assert pg is not None
        assert isinstance(pg, (dxtbx.model.Detector, dxtbx.model.DetectorNode))
        p = pg.add_panel()
        p.set_type("SENSOR_PAD")
        p.set_name(name)
        p.set_local_frame(fast_axis, slow_axis, origin)
        p.set_pixel_size(pixel_size)
        p.set_image_size(image_size)
        p.set_trusted_range(trusted_range)
        p.set_thickness(thickness)
        p.set_material(material)
        p.set_mu(mu)
        p.set_px_mm_strategy(px_mm)

        if nxdata and nxdata.data_scale_factor and not nxdata.data_scale_factor.shape:
            p.set_gain(1 / nxdata.data_scale_factor)

    return detector


def get_detector_module_slices(
    nxdetector: nxmx.NXdetector,
) -> tuple[tuple[slice, ...], ...]:
    """Return the slices pointing to the hyperslab of data for each module.

    This will be a tuple of tuples, where each tuple contains the slices corresponding
    to the slow and fast dimensions respectively.
    """
    return tuple(
        tuple(
            slice(int(start), int(start + step), 1)
            for start, step in zip(module.data_origin, module.data_size)
        )
        for module in nxdetector.modules
    )


def get_static_mask(nxdetector: nxmx.NXdetector) -> tuple[flex.bool, ...] | None:
    """Return the static mask for an NXdetector.

    This will be a tuple of flex.bool, of length equal to the number of modules. The
    result is intended to be compatible with the get_static_mask() method of dxtbx
    format classes.
    """
    try:
        pixel_mask = nxdetector.pixel_mask
    except KeyError:
        return None
    if pixel_mask is None or not pixel_mask.size:
        return None
    all_slices = get_detector_module_slices(nxdetector)
    all_mask_slices = []
    for slices in all_slices:
        mask_slice = flumpy.from_numpy(np.ascontiguousarray(pixel_mask[slices])) == 0
        mask_slice.reshape(
            flex.grid(mask_slice.all()[-2:])
        )  # handle 3 or 4 dimension arrays
        all_mask_slices.append(mask_slice)
    return tuple(all_mask_slices)


def _dataset_as_flex(
    data: h5py.Dataset,
    slices: tuple[slice, ...] | None,
    bit_depth: Literal[32] | None = None,
) -> flex.float | flex.double | flex.int:
    """
    Convert an HDF5 dataset to one of the expected flex types.

    Args:
        data: The HDF5 Dataset to convert
        slices:
            The Dataset will be sliced and made contiguous to this shape.
        bit_depth:
            If set to 32, and the dataset in signed integer
            representation requires more than 32-bit, then the data will
            be converted to a signed 32-bit integer. Without this, such
            attempted conversions will raise a TypeError.
    """
    # Make a guaranteed-contiguous copy of the sliced data
    data_np = np.ascontiguousarray(data[slices or ()])
    dtype = data_np.dtype

    # Handle integer conversion. Safe to convert if:
    # - Is signed and <= 4 bytes
    # - Is unsigned and <= 2 bytes
    #
    # Unsafe conversions to 32-bit integer can occur, but only if
    # bit_depth is explicitly set to 32.
    if np.issubdtype(dtype, np.integer):
        if (
            (np.issubdtype(dtype, np.signedinteger) and dtype.itemsize <= 4)
            or (np.issubdtype(dtype, np.unsignedinteger) and dtype.itemsize <= 2)
            or bit_depth == 32
        ):
            data_np = data_np.astype(np.int32, copy=False)
        else:
            raise TypeError(f"Unsupported integer dtype {data_np.dtype}")
    elif np.issubdtype(dtype, np.floating):
        if dtype.itemsize <= 4:
            # Promote anything <= single precision up to single precision
            data_np = data_np.astype(np.float32, copy=False)
        else:
            # Otherwise, everything else becomes double
            data_np = data_np.astype(np.float64, copy=False)
    else:
        # Isn't a recognised integer or floating point type
        raise TypeError(f"Unsupported dtype {data_np.dtype}")

    data_flex = flumpy.from_numpy(data_np)
    return data_flex


def get_raw_data(
    nxdata: nxmx.NXdata,
    nxdetector: nxmx.NXdetector,
    index: int,
    bit_depth: int | None = None,
) -> tuple[flex.float | flex.double | flex.int, ...]:
    """Return the raw data for an NXdetector.

    This will be a tuple of flex.float, flex.double or flex.int arrays, of length equal
    to the number of modules. The result is intended to be compatible with the
    get_raw_data() method of dxtbx format classes.
    """
    if nxdata.signal:
        try:
            data = nxdata[nxdata.signal]
        except KeyError:
            logger.warning(f"Key {nxdata.signal} specified by NXdata.signal missing")
            data = list(nxdata.values())[0]
    else:
        data = list(nxdata.values())[0]
    all_data = []
    sliced_outer = data[index]
    for module_slices in get_detector_module_slices(nxdetector):
        data_as_flex = _dataset_as_flex(
            sliced_outer, tuple(module_slices), bit_depth=bit_depth
        )
        data_as_flex.reshape(
            flex.grid(data_as_flex.all()[-2:])
        )  # handle 3 or 4 dimension arrays
        all_data.append(data_as_flex)
    return tuple(all_data)



"""Tests for dxbtx.format.FormatNXmx format classes."""

from __future__ import annotations

import h5py
import numpy as np
import pytest

from dxtbx.format.FormatNXmx import FormatNXmx
from dxtbx.format.FormatNXmxDLSI19_2 import FormatNXmxDLSI19_2
from dxtbx.model.experiment_list import ExperimentListFactory


@pytest.fixture
def nxmx_example_on_disk(tmp_path, nxmx_example):
    """
    Copy the in-memory example NXmx data to a file on disk.

    Since ExperimentListFactory.from_filenames can only read from disk, and requires
    the presence of a pixel mask, create a modified on-disk copy of the NXmx example
    data.
    """
    with nxmx_example as f, h5py.File(tmp_path / "test.h5", "w") as g:
        f.copy("entry", g)

        # Define a pixel mask, as required for creating a DXTBX model.
        pixel_mask = "/entry/instrument/detector/pixel_mask"
        g.create_dataset(pixel_mask, data=np.zeros((4362, 4148), dtype=np.int32))

        yield g


parameters = [("DIAMOND BEAMLINE I19-2", FormatNXmxDLSI19_2), ("DIAD", FormatNXmx)]


@pytest.mark.parametrize("instrument, format_class", parameters, ids=["I19-2", "DIAD"])
def test_Format_NXmx(nxmx_example_on_disk, instrument, format_class):
    """Check the right format class is used for the specified instrument."""
    name = "/entry/instrument/name"
    with nxmx_example_on_disk as g:
        del g[name]
        g.create_dataset(name, data=instrument)
        filename = g.filename
    (experiment,) = ExperimentListFactory.from_filenames([filename])
    assert experiment.imageset.get_format_class() == format_class



from __future__ import annotations

import copy
import pickle
import shutil
from unittest import mock

import pytest

from scitbx.array_family import flex

import dxtbx.format.FormatHDF5SaclaMPCCD
import dxtbx.format.image
import dxtbx.format.Registry
from dxtbx.format.FormatCBFMiniPilatus import FormatCBFMiniPilatus as FormatClass
from dxtbx.imageset import ExternalLookup, ImageSequence, ImageSetData, ImageSetFactory
from dxtbx.model import Beam, Detector, Panel
from dxtbx.model.beam import BeamFactory
from dxtbx.model.experiment_list import ExperimentListFactory

from . import imagelist


@pytest.mark.parametrize(
    "indices,expected_call_count,lazy",
    ((None, 4, False), ([1], 2, False), (None, 1, True), ([1], 1, True)),
)
def test_single_file_indices(indices, expected_call_count, lazy, dials_data):
    def dummy_beam():
        return BeamFactory.simple(1.0)

    with mock.patch.object(
        dxtbx.format.FormatHDF5SaclaMPCCD.FormatHDF5SaclaMPCCD,
        "_beam",
        side_effect=dummy_beam,
    ) as obj:
        filename = dials_data("image_examples") / "SACLA-MPCCD-run266702-0-subset.h5"
        format_class = dxtbx.format.Registry.get_format_class_for_file(filename)
        iset = format_class.get_imageset(
            [filename], single_file_indices=indices, format_kwargs={"lazy": lazy}
        )
        assert obj.call_count == expected_call_count
        iset.reader().nullify_format_instance()


# parametrize with the dials_data fixture
@pytest.mark.parametrize("image", imagelist.image_examples)
def test_format(image):
    """Test that we can read examples of various image formats"""
    format_class = dxtbx.format.Registry.get_format_class_for_file(image)
    reader = format_class.get_reader()([image])

    N = len(reader)

    for i in range(N):
        reader.read(i)

    assert format_class.get_imageset([image])


def test_image_tile():
    data = flex.int(flex.grid(10, 10))
    name = "TileName"

    tile = dxtbx.format.image.ImageTileInt(data, name)

    assert tile.data().all_eq(data)
    assert tile.name() == name
    assert tile.empty() is False


def test_image():
    data = flex.int(flex.grid(10, 10))
    name = "TileName0"
    tile0 = dxtbx.format.image.ImageTileInt(data, name)
    image = dxtbx.format.image.ImageInt(tile0)
    for i in range(1, 4):
        data = flex.int(flex.grid(10, 10))
        name = "TileName%d" % i
        tile = dxtbx.format.image.ImageTileInt(data, name)
        image.append(tile)

    assert image.n_tiles() == 4
    for i in range(image.n_tiles()):
        tile = image.tile(i)
        assert tile.name() == "TileName%d" % i


def test_image_buffer():
    data = flex.int(flex.grid(10, 10))
    name = "TileName0"
    tile0 = dxtbx.format.image.ImageTileInt(data, name)
    image = dxtbx.format.image.ImageInt(tile0)

    b = dxtbx.format.image.ImageBuffer(image)
    assert b.is_int() is True
    assert b.is_float() is False
    assert b.is_double() is False
    assert b.is_empty() is False


def test_external_lookup():
    mask = flex.bool(flex.grid(10, 10), True)
    gain = flex.double(flex.grid(10, 10), 1)
    pedestal = flex.double(flex.grid(10, 10), 2)

    lookup = ExternalLookup()
    lookup.mask.data = dxtbx.format.image.ImageBool(
        dxtbx.format.image.ImageTileBool(mask)
    )
    lookup.gain.data = dxtbx.format.image.ImageDouble(
        dxtbx.format.image.ImageTileDouble(gain)
    )
    lookup.pedestal.data = dxtbx.format.image.ImageDouble(
        dxtbx.format.image.ImageTileDouble(pedestal)
    )

    mask2 = lookup.mask.data.tile(0).data()
    gain2 = lookup.gain.data.tile(0).data()
    pedestal2 = lookup.pedestal.data.tile(0).data()

    assert mask2.all_eq(mask)
    assert gain2.all_eq(gain)
    assert pedestal2.all_eq(pedestal)


def test_imagesetdata(centroid_files):
    ReaderClass = FormatClass.get_reader()

    reader = ReaderClass(centroid_files)
    masker = FormatClass(centroid_files[0]).get_masker()

    handle = ImageSetData(reader, masker)

    assert handle.get_data(0).as_int().tile(0).data()

    assert handle.has_single_file_reader() is False

    path = handle.get_path(0)
    assert path == centroid_files[0]

    master_path = handle.get_master_path()
    assert master_path == ""

    identifier = handle.get_image_identifier(0)
    assert identifier == centroid_files[0]

    beam = FormatClass(centroid_files[0]).get_beam()
    detector = FormatClass(centroid_files[0]).get_detector()
    goniometer = FormatClass(centroid_files[0]).get_goniometer()
    scan = FormatClass(centroid_files[0]).get_scan()

    handle.set_beam(beam, 0)
    handle.set_detector(detector, 0)
    handle.set_goniometer(goniometer, 0)
    handle.set_scan(scan, 0)

    beam2 = handle.get_beam(0)
    detector2 = handle.get_detector(0)
    goniometer2 = handle.get_goniometer(0)
    scan2 = handle.get_scan(0)

    assert beam2 == beam
    assert detector2 == detector
    assert goniometer2 == goniometer
    assert scan2 == scan

    mask = flex.bool(flex.grid(10, 10), True)
    gain = flex.double(flex.grid(10, 10), 1)
    pedestal = flex.double(flex.grid(10, 10), 2)

    handle.external_lookup.mask.data = dxtbx.format.image.ImageBool(
        dxtbx.format.image.ImageTileBool(mask)
    )
    handle.external_lookup.gain.data = dxtbx.format.image.ImageDouble(
        dxtbx.format.image.ImageTileDouble(gain)
    )
    handle.external_lookup.pedestal.data = dxtbx.format.image.ImageDouble(
        dxtbx.format.image.ImageTileDouble(pedestal)
    )

    mask2 = handle.external_lookup.mask.data.tile(0).data()
    gain2 = handle.external_lookup.gain.data.tile(0).data()
    pedestal2 = handle.external_lookup.pedestal.data.tile(0).data()

    assert mask2.all_eq(mask)
    assert gain2.all_eq(gain)
    assert pedestal2.all_eq(pedestal)


@pytest.fixture(scope="session")
def centroid_files(dials_data):
    return [
        str(dials_data("centroid_test_data") / f"centroid_{i:04d}.cbf")
        for i in range(1, 10)
    ]


@pytest.fixture
def centroid_files_and_imageset(centroid_files):
    # Create the format class
    format_class = dxtbx.format.Registry.get_format_class_for_file(centroid_files[0])

    # Create the reader
    imageset = format_class.get_imageset(centroid_files, as_imageset=True)

    return centroid_files, imageset


def assert_is_iterable(iterator):
    list(iterator)


def assert_can_get_detectorbase(obj, indices, outside_index):
    for i in indices:
        obj.get_detectorbase(i)

    with pytest.raises(RuntimeError):
        obj.get_detectorbase(outside_index)


class TestImageSet:
    def test_imageset(self, centroid_files_and_imageset):
        filenames, imageset = centroid_files_and_imageset

        # Run a load of tests
        self.tst_get_item(imageset)
        assert len(imageset) == len(filenames)
        assert_is_iterable(imageset)
        self.tst_paths(imageset, filenames)
        assert_can_get_detectorbase(imageset, range(len(filenames)), 9)
        self.tst_get_models(imageset, list(range(len(filenames))), 9)

    def tst_get_item(self, imageset):
        assert imageset[0]
        with pytest.raises(RuntimeError):
            _ = imageset[9]

        imageset2 = imageset[3:7]
        assert imageset2[0]
        with pytest.raises(RuntimeError):
            _ = imageset2[5]

        assert len(imageset2) == 4
        assert_can_get_detectorbase(imageset2, range(0, 4), 5)
        self.tst_get_models(imageset2, range(0, 4), 5)
        self.tst_paths(imageset2, imageset.paths()[3:7])
        assert_is_iterable(imageset2)

        imageset2 = imageset[3:5]
        assert imageset2[0]
        with pytest.raises(RuntimeError):
            _ = imageset2[2]

        assert len(imageset2) == 2
        assert_can_get_detectorbase(imageset2, range(0, 2), 2)
        self.tst_get_models(imageset2, range(0, 2), 2)
        self.tst_paths(imageset2, imageset.paths()[3:5])
        assert_is_iterable(imageset2)

    @staticmethod
    def tst_paths(imageset, filenames1):
        filenames2 = imageset.paths()
        for f1, f2 in zip(filenames1, filenames2):
            assert f1 == f2

    def tst_get_models(self, imageset, indices, outside_index):
        for i in indices:
            self.tst_get_models_index(imageset, i)

        with pytest.raises(RuntimeError):
            self.tst_get_models_index(imageset, outside_index)

    @staticmethod
    def tst_get_models_index(imageset, index=None):
        imageset.get_detector(index)
        imageset.get_beam(index)

    @staticmethod
    def tst_set_models(imageset):
        # Create some other models
        beam = Beam((1, 0, 0), 0.5)
        detector = Detector(
            Panel(
                "UNKNOWN",
                "Panel",
                (1, 0, 0),
                (0, 1, 0),
                (0, 0, 1),
                (0.1, 0.1),
                (1000, 1000),
                (0, 1),
            )
        )

        # Override sequence models
        imageset.set_beam(beam)
        imageset.set_detector(detector)

        # Ensure this doesn't interfere with reading
        for i in imageset:
            pass

        # Get the models back and check they're ok
        beam2 = imageset.get_beam()
        detector2 = imageset.get_detector()
        assert beam2 == beam
        assert detector2 == detector

        # Get the models from an index back and check they're not the same
        beam2 = imageset.get_beam(0)
        detector2 = imageset.get_detector(0)
        assert beam2 != beam
        assert detector2 != detector


class TestImageSequence:
    def test(self, centroid_files):
        # Create the format class
        format_class = dxtbx.format.Registry.get_format_class_for_file(
            centroid_files[0]
        )

        # Create the sequence
        sequence = format_class.get_imageset(centroid_files)

        # Run a load of tests
        assert len(sequence) == len(centroid_files)
        assert sequence.get_array_range() == (0, 9)
        self.tst_get_item(sequence)
        assert_is_iterable(sequence)
        self.tst_paths(sequence, centroid_files)
        assert_can_get_detectorbase(sequence, range(len(centroid_files)), 9)
        self.tst_get_models(sequence, range(len(centroid_files)), 9)
        self.tst_set_models(sequence)

    def tst_get_item(self, sequence):
        _ = sequence[0]
        with pytest.raises(RuntimeError):
            _ = sequence[9]

        sequence2 = sequence[3:7]
        assert sequence2.get_array_range() == (3, 7)
        _ = sequence2[0]
        with pytest.raises(RuntimeError):
            _ = sequence2[5]

        # Check data access matches expected images from the slice
        panel_data1 = sequence[3][0]
        panel_data2 = sequence2[0][0]
        assert panel_data1.all_eq(panel_data2)

        assert len(sequence2) == 4
        assert_can_get_detectorbase(sequence2, range(0, 4), 5)
        self.tst_get_models(sequence2, range(0, 4), 5)
        self.tst_paths(sequence2, sequence.paths()[3:7])
        assert_is_iterable(sequence2)

        with pytest.raises(IndexError):
            _ = sequence[3:7:2]

        # Batch offset should not affect slicing of imagesequence
        # Simulate a scan starting from image 0
        sequence_ = copy.deepcopy(sequence)
        sequence_.get_scan().set_batch_offset(-1)
        sequence3 = sequence_[3:7]
        assert sequence3.get_array_range() == (3, 7)

    @staticmethod
    def tst_paths(sequence, filenames1):
        filenames2 = sequence.paths()
        for f1, f2 in zip(filenames1, filenames2):
            assert f1 == f2

    def tst_get_models(self, sequence, indices, outside_index):
        self.tst_get_models_index(sequence)
        for i in indices:
            self.tst_get_models_index(sequence, i)

    @staticmethod
    def tst_get_models_index(sequence, index=None):
        if index is None:
            sequence.get_detector()
            sequence.get_beam()
            sequence.get_goniometer()
            sequence.get_scan()
        else:
            sequence.get_detector(index)
            sequence.get_beam(index)
            sequence.get_goniometer(index)
            sequence.get_scan(index)

        # Ensure state at zero
        sequence[0]
        scan1 = sequence.get_scan()
        # Put sequence to end
        sequence[len(sequence) - 1]
        scan2 = sequence.get_scan()
        assert scan1 == scan2

    @staticmethod
    def tst_set_models(sequence):
        # Get some models
        beam = sequence.get_beam()
        gonio = sequence.get_goniometer()
        detector = sequence.get_detector()

        # Modify the geometry
        assert len(detector) == 1
        beam.set_direction((1, 0, 0))
        gonio.set_rotation_axis((0, 1, 0))
        detector[0].set_local_frame((1, 0, 0), (0, 1, 0), (0, 0, 1))

        # Override sequence models
        sequence.set_beam(beam)
        sequence.set_goniometer(gonio)
        sequence.set_detector(detector)

        # Ensure this doesn't interfere with reading
        for i in sequence:
            pass

        # Get the models back and check they're ok
        beam2 = sequence.get_beam()
        gonio2 = sequence.get_goniometer()
        detector2 = sequence.get_detector()
        assert beam2 == beam
        assert gonio2 == gonio
        assert detector2 == detector

        # Get the models from an index back and check they're the same
        beam2 = sequence.get_beam(0)
        gonio2 = sequence.get_goniometer(0)
        detector2 = sequence.get_detector(0)
        assert beam2 == beam
        assert gonio2 == gonio
        assert detector2 == detector

        # Get a sub sequence
        sub_sequence = sequence[3:7]

        # Get the models back and check they're ok
        beam2 = sub_sequence.get_beam()
        gonio2 = sub_sequence.get_goniometer()
        detector2 = sub_sequence.get_detector()
        assert beam2 == beam
        assert gonio2 == gonio
        assert detector2 == detector

        # Get the models from an index back and check they're not the same
        beam2 = sub_sequence.get_beam(0)
        gonio2 = sub_sequence.get_goniometer(0)
        detector2 = sub_sequence.get_detector(0)
        assert beam2 == beam
        assert gonio2 == gonio
        assert detector2 == detector


@pytest.mark.parametrize("lazy", (True, False))
def test_SACLA_MPCCD_Cheetah_File(dials_data, lazy):
    pytest.importorskip("h5py")
    filename = dials_data("image_examples") / "SACLA-MPCCD-run266702-0-subset.h5"

    format_class = dxtbx.format.Registry.get_format_class_for_file(filename)

    iset = format_class.get_imageset([filename], format_kwargs={"lazy": lazy})

    assert len(iset) == 4
    for i in range(len(iset)):
        assert iset.get_raw_data(i)
        #   assert iset.get_mask(i)
        assert iset.get_beam(i)
        assert iset.get_detector(i)
        assert iset.get_goniometer(i) is None
        assert iset.get_scan(i) is None

    iset = format_class.get_imageset(
        [filename], single_file_indices=[1], format_kwargs={"lazy": lazy}
    )
    assert len(iset) == 1

    for i in range(len(iset)):
        assert iset.get_raw_data(i)
        #   assert iset.get_mask(i)
        assert iset.get_beam(i)
        assert iset.get_detector(i)
        assert iset.get_goniometer(i) is None
        assert iset.get_scan(i) is None
    iset.reader().nullify_format_instance()


def test_imagesetfactory(centroid_files, dials_data):
    sequence = ImageSetFactory.new(centroid_files)

    assert isinstance(sequence[0], ImageSequence)

    template = str(dials_data("centroid_test_data") / "centroid_####.cbf")
    image_range = (3, 6)

    sequence = ImageSetFactory.from_template(template, image_range)

    assert isinstance(sequence[0], ImageSequence)
    assert len(sequence[0]) == 4
    assert sequence[0].paths()[0].endswith("3.cbf")
    assert sequence[0].paths()[-1].endswith("6.cbf")

    imageset = ImageSetFactory.make_imageset(centroid_files)
    assert len(imageset) == 9

    imageset = ImageSetFactory.make_imageset(centroid_files, check_format=False)
    assert len(imageset) == 9

    sequence = ImageSetFactory.make_sequence(template, list(range(1, 9 + 1)))
    assert len(sequence) == 9

    sequence = ImageSetFactory.make_sequence(template, list(range(3, 6 + 1)))
    assert len(sequence) == 4


def test_make_sequence_with_percent_character(dials_data, tmp_path):
    images = [
        dials_data("centroid_test_data") / f"centroid_{i:04}.cbf" for i in range(1, 10)
    ]
    directory = tmp_path / "test%"
    directory.mkdir()
    try:
        for image in images:
            try:
                (directory / image.name).symlink_to(image)
            except OSError:
                shutil.copy(image, directory)

        template = str(directory / "centroid_####.cbf")
        sequence = ImageSetFactory.make_sequence(template, range(1, 10))
        assert len(sequence) == 9

        sequences = ImageSetFactory.new(
            [str(directory / image.name) for image in images]
        )
        assert len(sequences) == 1
        assert len(sequences[0]) == 9

        sequences = ImageSetFactory.from_template(template)
        assert len(sequences) == 1
        assert len(sequences[0]) == 9

    finally:  # clean up potentially copied files after running test
        try:
            # Force the dxtbx filehandler cache to close any open handles
            sequences[0].get_format_class().get_cache_controller().check(
                None, lambda: None
            )
        except Exception:
            pass
        for image in images:
            try:
                (directory / image.name).unlink()
            except (FileNotFoundError, PermissionError):
                pass


def test_pickle_imageset(centroid_files):
    sequence = ImageSetFactory.new(centroid_files)[0]

    # Read the 5th image
    assert sequence[4]

    # Set a rejected image
    assert not sequence.is_marked_for_rejection(8)
    sequence.mark_for_rejection(8, True)
    assert sequence.is_marked_for_rejection(8)

    # Pickle, then unpickle
    pickled_sequence = pickle.dumps(sequence)
    sequence2 = pickle.loads(pickled_sequence)

    assert sequence.get_template() == sequence2.get_template()
    assert sequence.get_array_range() == sequence2.get_array_range()
    assert sequence.get_beam() == sequence2.get_beam()
    assert sequence.get_goniometer() == sequence2.get_goniometer()
    assert sequence.get_scan() == sequence2.get_scan()
    assert sequence.paths() == sequence2.paths()
    assert sequence == sequence2

    # Check auxiliary methods after pickling
    sequence3 = sequence2[0:4]
    sequence4 = sequence3[0:2]
    sequence4.get_detectorbase(0)
    sequence4[0]

    # Check reject list is preserved for https://github.com/dials/dials/issues/2998
    assert sequence2.is_marked_for_rejection(8)


def test_get_corrected_data(centroid_files):
    sequence = ImageSetFactory.new(centroid_files)[0]

    assert not sequence.get_gain(0)
    assert not sequence.get_pedestal(0)
    data1 = sequence.get_corrected_data(0)[0]

    # Set a gain
    detector = sequence.get_detector()
    panel = detector[0]
    panel.set_gain(2)
    data2 = sequence.get_corrected_data(0)[0]
    assert flex.mean(data2) == pytest.approx(flex.mean(data1) / 2.0)

    # Set a pedestal
    panel.set_pedestal(1)
    data3 = sequence.get_corrected_data(0)[0]
    assert flex.mean(data3) == pytest.approx(flex.mean(data2) - 1.0 / 2.0)


def test_multi_panel_gain_map(dials_data):
    pytest.importorskip("h5py")
    filename = dials_data("image_examples") / "SACLA-MPCCD-run266702-0-subset.h5"

    format_class = dxtbx.format.Registry.get_format_class_for_file(filename)
    iset = format_class.get_imageset([filename])

    # Test gain map set correctly (https://github.com/dials/dials/issues/979)
    gain_values = [p.get_gain() for p in iset.get_detector(0)]
    gain_maps = iset.get_gain(0)
    for v, m in zip(gain_values, gain_maps):
        assert m.all_eq(v)

    iset.reader().nullify_format_instance()


@pytest.mark.parametrize(
    "multi_panel,expected_panel_count",
    (
        (False, 24),
        (True, 120),
    ),
)
def test_multi_panel(multi_panel, expected_panel_count, dials_data):
    image_path = str(dials_data("image_examples") / "DLS_I23-germ_13KeV_0001.cbf.bz2")
    experiments = ExperimentListFactory.from_filenames(
        [image_path], format_kwargs={"multi_panel": multi_panel}
    )
    imageset = experiments[0].imageset
    assert (
        len(imageset.get_detector())
        == len(imageset.get_raw_data(0))
        == expected_panel_count
    )


def test_scan_imageset_slice_consistency(dials_data):
    files = sorted(dials_data("centroid_test_data").glob("*.cbf"))[1:]
    expt = ExperimentListFactory.from_filenames(str(f) for f in files)[0]
    assert expt.scan[0:8] == expt.scan
    # The following doesn't work, and expects expt.imageset[1:9]
    assert expt.imageset[0:8] == expt.imageset



from __future__ import annotations

import contextlib
import importlib
import inspect
import io
import os
import subprocess
import sys
from pathlib import Path

import libtbx
import libtbx.pkg_utils

try:
    import dials.precommitbx.nagger

    dials.precommitbx.nagger.nag()
except ModuleNotFoundError:
    pass


def _find_site_packages_with_metadata(package_name: str, build_path: Path):
    """
    Find the site-packages directory containing the package metadata.
    Returns the site-packages directory if metadata is found, None otherwise.
    """
    # Look for Python site-packages directories in the build path
    for python_dir in build_path.glob("lib/python*"):
        site_packages = python_dir / "site-packages"
        if site_packages.exists():
            # Look for both .dist-info and .egg-info directories
            for pattern in [f"{package_name}*.dist-info", f"{package_name}*.egg-info"]:
                for metadata_dir in site_packages.glob(pattern):
                    if metadata_dir.exists():
                        # Return site-packages only if we actually found metadata
                        return site_packages
            # If no metadata found in this site-packages, continue searching
    return None


def _install_setup_readonly_fallback(package_name: str):
    """
    Partially install package in the libtbx build folder.

    This is a less complete installation - base python console_scripts
    entrypoints will not be installed, but the basic package metadata
    and other entrypoints will be enumerable through dispatcher black magic
    """
    root_path = libtbx.env.dist_path(package_name)
    import_path = os.path.join(root_path, "src")

    # Install this into a build/dxtbx subfolder
    build_path = abs(libtbx.env.build_path / package_name)
    subprocess.run(
        [
            sys.executable,
            "-m",
            "pip",
            "install",
            "--prefix",
            build_path,
            "--no-build-isolation",
            "--no-deps",
            "-e",
            root_path,
        ],
        check=True,
    )

    # Get the actual environment being configured (NOT libtbx.env)
    env = _get_real_env_hack_hack_hack()

    # As of PEP 660, the package metadata (dist-info) goes in the install dir,
    # not the source dir. Add this location to the python path.
    metadata_dir = _find_site_packages_with_metadata(package_name, Path(build_path))
    if metadata_dir and metadata_dir not in sys.path:
        metadata_rel = libtbx.env.as_relocatable_path(str(metadata_dir))
        if metadata_rel not in env.pythonpath:
            env.pythonpath.insert(0, metadata_rel)

    # Update the sys.path so we can find the package in this process
    # if we do a full reconstruction of the working set
    if import_path not in sys.path:
        sys.path.insert(0, import_path)

    # ...and make sure it is picked up by the import system
    importlib.invalidate_caches()

    # This is already generated by this point, but will get picked up
    # on the second libtbx.refresh.
    module = env.module_dict[package_name]
    if f"src/{package_name}" not in module.extra_command_line_locations:
        module.extra_command_line_locations.append(f"src/{package_name}")

    # Because dispatchers for all modules are generated _before_ any of
    # libtbx_refresh are run, then we need to regenerate all of the
    # dispatchers now we've added the extra PYTHONPATH
    with contextlib.redirect_stdout(io.StringIO()):
        for module in env.module_list:
            module.process_command_line_directories()


def _get_real_env_hack_hack_hack():
    """
    Get the real, currently-being-configured libtbx.env environment.

    This is not libtbx.env, because although libtbx.env_config.environment.cold_start
    does:
        self.pickle()
        libtbx.env = self
    the first time there is an "import libtbx.load_env" this environment
    gets replaced by unpickling the freshly-written libtbx_env file onto
    libtbx.env, thereby making the environment accessed via libtbx.env
    *not* the actual one that is currently being constructed.

    So, the only way to get this environment being configured in order
    to - like - configure it, is to walk the stack trace and extract the
    self object from environment.refresh directly.
    """
    for frame in inspect.stack():
        if (
            frame.filename.endswith("env_config.py")
            and frame.function == "refresh"
            and "self" in frame.frame.f_locals
        ):
            return frame.frame.f_locals["self"]

    raise RuntimeError("Could not determine real libtbx.env_config.environment object")


# When building in libtbx, always assume it's unsafe to write to base/
_install_setup_readonly_fallback("dxtbx")



from __future__ import annotations

import os

import pytest

from dxtbx.format.FormatNXmxDLS16MI03 import FormatNXmxDLS16MI03
from dxtbx.model.experiment_list import ExperimentListFactory

dials = pytest.importorskip("dials")
pytest.importorskip("h5py")


@pytest.mark.skip(reason="Data erased from CM visit")
@pytest.mark.parametrize(
    "master_h5,masked_count",
    [
        (
            "/dls/i03/data/2021/cm28170-2/TestProteinaseK/protk_4/protk_4_5.nxs",
            1052672,
        ),  # full detector mode
        (
            "/dls/i03/data/2021/cm28170-2/xraycentring/TestProteinaseK/protk_5/protk_5_15.nxs",
            526336,
        ),  # ROI mode
    ],
)
def test_masked_i03(master_h5, masked_count):
    if not os.access(master_h5, os.R_OK):
        pytest.skip("Test images not available")

    assert FormatNXmxDLS16MI03.understand(master_h5)
    expts = ExperimentListFactory.from_filenames([master_h5])
    assert (
        expts[0].detector[0].get_untrusted_rectangle_mask().count(False) == masked_count
    )



from __future__ import annotations

import copy
import os
import pickle
import sys

from iotbx.detectors.cspad_detector_formats import (
    detector_format_version,
    reverse_timestamp,
)
from iotbx.detectors.npy import NpyImage, image_dict_to_unicode
from scitbx.array_family import flex
from spotfinder.applications.xfel import cxi_phil

from dxtbx.format.FormatPY import FormatPY


class FormatPYunspecified(FormatPY):
    @staticmethod
    def understand(image_file):
        """Seems like the static method wastes a lot of effort here; it's not possible to
        just read the first few bytes; instead understand() reads the entire first data
        item in the file; an entire binary image.  This data is then read again in the
        _start() method and again in the detectorbase constructor."""
        try:
            with FormatPYunspecified.open_file(image_file, "rb") as fh:
                data = pickle.load(fh, encoding="bytes")
                data = image_dict_to_unicode(data)
        except (AttributeError, OSError):
            return False

        headers = set(data)
        wanted_header_items = {"SIZE1", "SIZE2", "TIMESTAMP"}

        return wanted_header_items.issubset(headers)

    def detectorbase_start(self):
        pass

    def _start(self):
        if isinstance(self._image_file, str) and os.path.isfile(self._image_file):
            with FormatPYunspecified.open_file(self._image_file, "rb") as fh:
                data = pickle.load(fh, encoding="bytes")
                data = image_dict_to_unicode(data)
        else:
            data = self._image_file

        if "DETECTOR_ADDRESS" not in data:
            # legacy format; try to guess the address
            self.LCLS_detector_address = "CxiDs1-0|Cspad-0"
            if "DISTANCE" in data and data["DISTANCE"] > 1000:
                # downstream CS-PAD detector station of CXI instrument
                self.LCLS_detector_address = "CxiDsd-0|Cspad-0"
        else:
            self.LCLS_detector_address = data["DETECTOR_ADDRESS"]

        self._timesec = reverse_timestamp(data["TIMESTAMP"])[0]

        version_lookup = detector_format_version(
            self.LCLS_detector_address, self._timesec
        )
        self.start_helper(
            version_token="distl.detector_format_version=%s" % version_lookup
        )

    def start_helper(self, version_token):
        is_file = isinstance(self._image_file, str) and os.path.isfile(self._image_file)

        if is_file:
            file_name = self._image_file
        else:
            file_name = "inmem"

        args = [
            file_name,
            version_token,
            "viewer.powder_arcs.show=False",
            "viewer.powder_arcs.code=3n9c",
        ]

        params = cxi_phil.cxi_versioned_extract(args)
        horizons_phil = params.persist.commands

        if is_file:
            image = NpyImage(file_name)
        else:
            print(
                "This is not a file; assume the data are in the defined dictionary format"
            )
            image = NpyImage(file_name, source_data=self._image_file)
        image.readHeader(horizons_phil)
        image.translate_tiles(horizons_phil)
        # necessary to keep the phil parameters for subsequent calls to get_tile_manager()
        image.horizons_phil_cache = copy.deepcopy(horizons_phil)
        self.detectorbase = image

    def _goniometer(self):
        return self._goniometer_factory.single_axis()

    def _detector(self):
        """Return a model for a simple detector"""
        trusted_range = (
            self.detectorbase.parameters.get("MIN_TRUSTED_VALUE", 0),
            self.detectorbase.saturation,
        )

        return self._detector_factory.simple(
            sensor="PAD",
            distance=self.detectorbase.distance,
            beam_centre=(self.detectorbase.beamx, self.detectorbase.beamy),
            fast_direction="+x",
            slow_direction="-y",
            pixel_size=(self.detectorbase.pixel_size, self.detectorbase.pixel_size),
            image_size=(self.detectorbase.size2, self.detectorbase.size1),
            trusted_range=trusted_range,
            mask=[],
        )

    def _beam(self):
        """Return a simple model for the beam."""

        return self._beam_factory.simple(self.detectorbase.wavelength)

    def _scan(self):
        """Return the scan information for this image."""

        if (
            self.detectorbase.osc_start is not None
            and self.detectorbase.deltaphi is not None
            and self.detectorbase.deltaphi > 0
        ):
            exposure_time = self.detectorbase.parameters.get("TIME", 1)
            osc_start = self.detectorbase.osc_start
            osc_range = self.detectorbase.deltaphi
            timestamp = self._timesec

            return self._scan_factory.single_file(
                self._image_file, exposure_time, osc_start, osc_range, timestamp
            )

        else:
            return self._scan_factory.make_scan(
                image_range=(1, 1),
                # femtosecond X-ray pulse, set this to a dummy argument
                exposure_times=[1.0],
                oscillation=(0.0, 0.0),
                epochs={1: self._timesec},
            )

    def get_static_mask(self):
        """Creates a mask merging untrusted pixels with active areas."""
        detector_base = self.detectorbase
        # get effective active area coordinates
        tile_manager = detector_base.get_tile_manager(detector_base.horizons_phil_cache)
        tiling = tile_manager.effective_tiling_as_flex_int(
            reapply_peripheral_margin=True
        )
        if tiling is None or len(tiling) == 0:
            return None

        # get the raw data to get the size of the mask
        data = self.get_raw_data()

        # set the mask to the same dimensions as the data
        mask = flex.bool(flex.grid(data.focus()))

        # set active areas to True so they are not masked
        for i in range(len(tiling) // 4):
            x1, y1, x2, y2 = tiling[4 * i : (4 * i) + 4]
            sub_array = flex.bool(flex.grid(x2 - x1, y2 - y1), True)
            mask.matrix_paste_block_in_place(sub_array, x1, y1)
        return mask


class FormatPYunspecifiedInMemory(FormatPYunspecified):
    """Overrides the Format object's init method to accept an image dictionary
    instead of a file name. Used with XFELs when it is desirable to never write
    a file to disk, but to process it only in memory.
    """

    @staticmethod
    def understand(image_file):
        """If it's an image dictionary, we understand this"""
        wanted_header_items = ["SIZE1", "SIZE2", "TIMESTAMP"]
        try:
            if any(
                header_item not in image_file for header_item in wanted_header_items
            ):
                return False

            return True

        except AttributeError:
            return False

    def __init__(self, data, **kwargs):
        """@param data In memory image dictionary, already initialized"""
        super().__init__(data, **kwargs)
        self._image_file = copy.deepcopy(data)


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatPYunspecified.understand(arg))



from __future__ import annotations

import contextlib
import datetime
import logging
from functools import cached_property
from pathlib import Path

import h5py
import numpy as np
import nxmx

from dxtbx.format.FormatNXmx import FormatNXmx

logger = logging.getLogger(__name__)


def get_bit_depth_from_meta(meta_file_name: Path) -> int:
    with h5py.File(meta_file_name) as f:
        with contextlib.suppress(KeyError):
            return int(f["/_dectris/bit_depth_image"][()])

        # This might be a very old meta file, that only had 'datatype'?
        if "datatype" in f:
            unique_vals = np.unique(f["datatype"])
            if not len(unique_vals) == 1:
                raise RuntimeError(
                    f"Error: Could not determine bit depth from metafile {meta_file_name} (multiple bit depths)"
                )
            return np.dtype(unique_vals[0]).itemsize * 8

    raise RuntimeError("Cannot determine metafile bit depth")


def find_meta_filename(master_like: Path) -> Path:
    """
    Find the path to the '..._meta.h5' file in the same directory as the master file.
    Args:
        master_like:  File path of the master HDF5 file.

    Returns:
        File path of the HDF5 metadata '..._meta.h5' file.

    Raises:
        RuntimeError: If no meta-file was found.
    """

    def _local_visit(name) -> Path | None:
        obj = f[name]
        if not hasattr(obj, "keys"):
            return None
        for k in obj.keys():
            kclass = obj.get(k, getlink=True, getclass=True)
            if kclass is h5py._hl.group.ExternalLink:
                kfile = master_like.parent / obj.get(k, getlink=True).filename
                if kfile.stem.endswith("meta"):
                    return kfile
        return None

    master_like = Path(master_like)
    with h5py.File(master_like) as f:
        meta_filename = f.visit(_local_visit)

    if meta_filename is None:
        # Try a fallback of looking for a file named the same but "meta.h5"
        look_for = master_like.with_name(
            master_like.stem.removesuffix("_master") + "_meta.h5"
        )
        if look_for.is_file():
            meta_filename = look_for
        else:
            raise RuntimeError(f"Could not find h5 meta file for {master_like}")

    return meta_filename


class FormatNXmxDLS(FormatNXmx):
    _cached_file_handle = None

    @staticmethod
    def understand(image_file):
        with h5py.File(image_file) as handle:
            name = nxmx.h5str(FormatNXmxDLS.get_instrument_name(handle))
            if name and name.lower() in ("i03", "i04", "i24", "vmxi"):
                return True
            if name and "ebic" in name.lower():
                return False
            if name and name.upper().startswith(("DIAMOND BEAMLINE", "DLS")):
                return True
        return False

    def __init__(self, image_file, **kwargs):
        """Initialise the image structure from the given file."""
        self._legacy = None
        super().__init__(image_file, **kwargs)

    def _start(self):
        super()._start()

        if self._legacy:
            # Get the bit depth from the meta.h5 in order to distinguish masked and
            # saturated pixels. Ideally we would get this from
            # /entry/instrument/detector/bit_depth_readout.
            # See https://jira.diamond.ac.uk/browse/MXGDA-3674
            try:
                self._bit_depth_readout = get_bit_depth_from_meta(self._meta)
            except RuntimeError:
                logger.warning(
                    "Could not determine bit depth of legacy image data: Falling back to default of 16"
                )
                self._bit_depth_readout = 16

    @cached_property
    def _meta(self) -> Path:
        return find_meta_filename(self._image_file)

    def _get_nxmx(self, fh: h5py.File):
        nxmx_obj = nxmx.NXmx(fh)
        nxentry = nxmx_obj.entries[0]

        nxdetector = nxentry.instruments[0].detectors[0]
        if nxdetector.underload_value is None:
            nxdetector.underload_value = 0

        if self._legacy is None:
            name = nxmx.h5str(FormatNXmx.get_instrument_name(fh))
            if nxentry.start_time and "I03" in name.upper():
                self._legacy = nxentry.start_time.replace(
                    tzinfo=None
                ) < datetime.datetime(2021, 9, 10, 13, 12, 0)
            elif nxentry.start_time and "I04" in name.upper():
                self._legacy = nxentry.start_time.replace(
                    tzinfo=None
                ) < datetime.datetime(2021, 9, 14, 15, 3, 0)
            elif nxentry.start_time and "VMXM" in name.upper():
                # don't have a time stamp for precisely when this happened to VMXm
                self._legacy = nxentry.start_time.replace(
                    tzinfo=None
                ) < datetime.datetime(2022, 1, 1, 1, 0, 0)
            elif "VMXI" in name.upper():
                # Until some point between July 2021 and  November 2021 the
                # data_size recorded in the VMXi NeXus files were reversed
                self._legacy = np.all(nxdetector.modules[0].data_size == (2068, 2162))

        if self._legacy:
            # data_size was reversed, see https://jira.diamond.ac.uk/browse/MXGDA-3676
            for module in nxdetector.modules:
                module.data_size = module.data_size[::-1]
            try:
                nxdetector.pixel_mask
            except KeyError:
                nxdetector.pixel_mask = None
            if (
                nxdetector.pixel_mask is not None
                and nxdetector.pixel_mask.shape
                != tuple(nxdetector.modules[0].data_size)
            ):
                # At some point the pixel mask was stored with the dimensions reversed -> ignore
                # https://jira.diamond.ac.uk/browse/MXGDA-3675
                nxdetector.pixel_mask = None
        return nxmx_obj



from __future__ import annotations

from dxtbx.format.Format import abstract
from dxtbx.format.FormatMultiImage import FormatMultiImage


@abstract
class FormatMultiImageLazy(FormatMultiImage):
    """
    Lazy version of FormatMultiImage that does not instantiate the models ahead of time.
    It creates an ImageSetLazy class and returns it. Saves time when image file contains
    too many images to setup before processing.
    """

    @classmethod
    def get_imageset(
        Class,
        filenames,
        beam=None,
        detector=None,
        goniometer=None,
        scan=None,
        as_sequence=False,
        as_imageset=False,
        single_file_indices=None,
        format_kwargs=None,
        template=None,
        check_format=True,
        lazy=True,
    ):
        return super().get_imageset(
            filenames=filenames,
            beam=beam,
            detector=detector,
            goniometer=goniometer,
            scan=scan,
            as_sequence=as_sequence,
            as_imageset=as_imageset,
            single_file_indices=single_file_indices,
            format_kwargs=format_kwargs,
            template=template,
            check_format=check_format,
            lazy=lazy,
        )



from __future__ import annotations

import math
import os
import sys

import h5py
import numpy as np

from cctbx import factor_ev_angstrom
from cctbx.eltbx import attenuation_coefficient
from scitbx import matrix
from scitbx.array_family import flex

from dxtbx.format.FormatHDF5 import FormatHDF5
from dxtbx.format.FormatStill import FormatStill
from dxtbx.format.nexus import h5str
from dxtbx.model import ParallaxCorrectedPxMmStrategy
from dxtbx.model.detector import Detector

# 151028: deepcopying this class causes crash in h5py
#         temporary fix by closing the file in every methods(!)
# 161003: updated to follow dxtbx changes
#         removed iotbx support, which was incomplete anyway
# 161005: get wavelength from the file
# 170929: read metadata for phase III and compact MPCCDs
# 171003: fix mask
# 180724: update 'understand' to exclude Rayonix data


class FormatHDF5SaclaMPCCD(FormatHDF5, FormatStill):
    """
    Class to handle multi-event HDF5 files from MPCCD
    preprocessed by Cheetah SFX pipeline at SACLA.
    To handle reassembled images from "DataConvert3 -reconst"
    (old pipeline), use FormatHDF5Sacla.
    To override metrology, use the following environmental variables.
       MPCCD_GEOMETRY, MPCCD_DISTANCE
       MPCCD_RECONST_MODE

    You can also specify reference_geometry in dials.stills_process.
    """

    @staticmethod
    def understand(image_file):
        with h5py.File(image_file, "r") as h5_handle:
            if "metadata/detector" in h5_handle:
                if "Rayonix" in h5str(h5_handle["metadata/detector"][()]):
                    return False

            for elem in h5_handle:
                if elem.startswith("tag-"):
                    return True
            return False

    def __init__(self, image_file, index=0, reconst_mode=False, **kwargs):
        self._raw_data = None
        self.index = index
        self.image_filename = image_file
        self.lazy = kwargs.get("lazy", True)
        super().__init__(image_file, **kwargs)

        self.PIXEL_SIZE = 50 / 1000  # 50 um
        self.RECONST_SIZE = 2398  # compatible with DataConvert3 -reconst mode

        # These hard-coded values can be overwritten
        # by MPCCD_GEOMETRY and MPCCD_DISTANCE
        #
        # These values can be retrieved from SACLA API.
        # Alternatively, you can get it from a CrystFEL geometry file by
        # awk '/corner_x/{x=50*$3} /corner_y/{y=50*$3; printf x","y","rot","}
        #      /\/ss/{rot=-atan2($3, $4)/3.141592*180}' input.geom

        # Default value for Phase I MPCCD
        self.distance = 50.0  # mm
        self.panel_origins = [
            (-1755.000000, 51711.000000, 0.000000),
            (-1711.000000, 24944.000000, 0.000000),
            (817.000000, -1808.000000, 0.000000),
            (812.000000, -28466.000000, 0.000000),
            (-792.000000, 28544.000000, 0.000000),
            (-781.000000, 1840.000000, 0.000000),
            (1650.000000, -24900.000000, 0.000000),
            (1655.000000, -51626.000000, 0.000000),
        ]  # um
        self.panel_rotations = [
            -89.906197,
            -89.915802,
            -89.980003,
            -89.929298,
            89.963097,
            89.880798,
            90.000000,
            90.029503,
        ]
        self.thickness = 0.050
        self.mask = None

        # Read metadata if possible
        self.read_metadata()

        # Override by environmental variables
        if "MPCCD_RECONST_MODE" in os.environ:
            reconst_mode = bool(os.environ["MPCCD_RECONST_MODE"])
        self.RECONST_MODE = reconst_mode
        self.RECONST_64 = (
            True  # Set False if you want to keep panels completely horizontal
        )
        # But this makes errors bigger.

        if "MPCCD_GEOMETRY" in os.environ:
            try:
                tmp = [float(i) for i in os.environ["MPCCD_GEOMETRY"].split(",")]
                if len(tmp) != 24:
                    raise OSError(
                        "Environment variable MPCCD_GEOMETRY must contain 24 comma-separated parts"
                    )
                for i in range(8):
                    self.panel_origins[i] = (-tmp[i * 3], tmp[i * 3 + 1], 0)
                    self.panel_rotations[i] = tmp[i * 3 + 2]
            except Exception as e:
                raise OSError(
                    f"Invalid MPCCD Geometry specified in environment variable MPCCD_GEOMETRY: {e}"
                )
        if "MPCCD_DISTANCE" in os.environ:
            self.distance = float(os.environ["MPCCD_DISTANCE"])

    def _start(self):
        h5_handle = h5py.File(self.image_filename, "r")

        self._images = sorted([tag for tag in h5_handle if tag.startswith("tag-")])
        self.tag = self._images[self.index]
        h5_handle.close()

    def read_metadata(self):
        h5_handle = h5py.File(self.image_filename, "r")

        if "metadata" not in h5_handle:
            return
        try:
            distance = h5_handle["metadata/distance_in_mm"][()]
            panel_rotations = h5_handle["metadata/angle_in_rad"][()]
            posx = h5_handle["metadata/posx_in_um"][()]
            posy = h5_handle["metadata/posy_in_um"][()]
            posz = h5_handle["metadata/posz_in_um"][()]
            panel_origins = list(zip(posx, posy, posz))
            sensor = h5_handle["metadata/sensor_id"][0]
            thickness = 0.050
            if sensor.startswith(b"MPCCD-8B"):
                thickness = 0.300  # Phase 3 sensor
            orig_mask = np.logical_not(h5_handle["metadata/pixelmask"][()])
            mask = self.split_panels(orig_mask, bool=True)
        except Exception:
            return
        self.distance = distance
        self.panel_rotations = panel_rotations
        self.panel_origins = panel_origins
        self.thickness = thickness
        self.mask = mask

    def get_image_file(self, index=None):
        return self.image_filename

    def set_index(self, index):
        assert index < len(self._images)

        self.index = index
        self.tag = self._images[self.index]
        self._raw_data = None

    def _detector(self, index=None):
        wavelength = self.get_beam(index).get_wavelength()

        table = attenuation_coefficient.get_table("Si")
        mu = table.mu_at_angstrom(wavelength) / 10.0
        px_mm = ParallaxCorrectedPxMmStrategy(mu, self.thickness)

        if self.RECONST_MODE:
            return self._detector_factory.simple(
                sensor="PAD",
                distance=self.distance,
                beam_centre=(
                    self.RECONST_SIZE / 2 * self.PIXEL_SIZE,
                    self.RECONST_SIZE / 2 * self.PIXEL_SIZE,
                ),
                fast_direction="-x",
                slow_direction="-y",
                pixel_size=(self.PIXEL_SIZE, self.PIXEL_SIZE),
                image_size=(self.RECONST_SIZE, self.RECONST_SIZE),
                trusted_range=(0, 65535),
                mask=[],
            )  # TODO: add gaps

        detector = Detector()
        root = detector.hierarchy()
        root.set_frame((-1, 0, 0), (0, 1, 0), (0, 0, -self.distance))

        for i in range(8):
            angle = math.pi * self.panel_rotations[i] / 180.0
            fast = matrix.col((math.cos(angle), math.sin(angle), 0))
            slow = matrix.col((-math.sin(angle), math.cos(angle), 0))

            origin = (
                matrix.col(
                    (
                        -self.panel_origins[i][0],
                        self.panel_origins[i][1],
                        self.panel_origins[i][2],
                    )
                )
                / 1000.0
            )
            p = root.add_panel()
            p.set_type("SENSOR_PAD")
            p.set_name("Panel%d" % i)
            p.set_image_size((512, 1024))
            p.set_trusted_range((0, 65535))
            p.set_pixel_size((self.PIXEL_SIZE, self.PIXEL_SIZE))
            p.set_thickness(self.thickness)
            p.set_local_frame(fast.elems, slow.elems, origin.elems)
            p.set_px_mm_strategy(px_mm)
            p.set_gain(10)

        return detector

    def _beam(self):
        h5_handle = h5py.File(self.image_filename, "r")
        eV = h5_handle[self.tag]["photon_energy_ev"][()]
        h5_handle.close()

        return self._beam_factory.simple(factor_ev_angstrom / eV)

    def get_num_images(self):
        return len(self._images)

    def split_panels(self, img, bool=False):
        tmp = []

        for i in range(8):
            xmin, ymin, xmax, ymax = 0, i * 1024, 512, (i + 1) * 1024
            # To avoid "numpy.ndarray instance is not contiguous"
            if bool:
                source = np.ascontiguousarray(img[ymin:ymax, xmin:xmax])
                tmp.append(flex.bool(source))
            else:
                source = np.ascontiguousarray(img[ymin:ymax, xmin:xmax], dtype=np.int32)
                tmp.append(flex.int(source))

        return tuple(tmp)

    def get_raw_data(self, index=None):
        if index is not None and self.index != index:
            self.set_index(index)

        if self._raw_data is None:
            if self.RECONST_MODE:
                self._raw_data = flex.int(self.reconst_image())

            else:
                h5_handle = h5py.File(self.image_filename, "r")

                data = h5_handle[self.tag]["data"][()]  # .astype(np.int32)
                # [()] forces conversion to ndarray
                # this is 8192x512 (slow/fast) tiled image
                h5_handle.close()

                self._raw_data = self.split_panels(data)

        return self._raw_data

    def get_active_areas(self):
        assert self.RECONST_MODE

        return self.active_areas

    def reconst_image(self):
        det = np.empty((self.RECONST_SIZE, self.RECONST_SIZE), dtype="int32")
        det.fill(-1)

        h5_handle = h5py.File(self.image_filename, "r")
        data = h5_handle[self.tag]["data"][()].astype(np.int32)
        h5_handle.close()

        self.active_areas = []

        for i in range(8):
            angle = math.pi * self.panel_rotations[i] / 180.0
            fast = matrix.col((math.cos(angle), math.sin(angle)))
            slow = matrix.col((-math.sin(angle), math.cos(angle)))
            origin = matrix.col(
                (
                    -self.panel_origins[i][0] / self.PIXEL_SIZE / 1000
                    + self.RECONST_SIZE / 2,
                    -self.panel_origins[i][1] / self.PIXEL_SIZE / 1000
                    + self.RECONST_SIZE / 2,
                )
            )

            if self.RECONST_64:
                size_fast = 256
                size_slow = 256

                for j in range(2):
                    for k in range(4):
                        xmin, ymin, xmax, ymax = (
                            j * size_slow,
                            (i * 4 + k) * size_fast,
                            (j + 1) * size_slow,
                            (i * 4 + k + 1) * size_fast,
                        )
                        source = data[ymin:ymax, xmin:xmax].transpose()

                        subpanel_origin = (
                            origin - j * size_fast * fast + k * size_slow * slow
                        )
                        if abs(round(self.panel_rotations[i]) + 90) < 1:
                            det[
                                int(round(subpanel_origin[1])) : int(
                                    round(subpanel_origin[1] + size_slow)
                                ),
                                int(round(subpanel_origin[0])) : int(
                                    round(subpanel_origin[0] + size_fast)
                                ),
                            ] = source
                            # TODO: Is the border inclusive?
                            self.active_areas.extend(
                                [
                                    round(subpanel_origin[1]),
                                    round(subpanel_origin[0]),
                                    round(subpanel_origin[1] + size_slow),
                                    round(subpanel_origin[0] + size_fast),
                                ]
                            )
                        elif abs(round(self.panel_rotations[i]) - 90) < 1:
                            det[
                                int(round(subpanel_origin[1])) : int(
                                    round(subpanel_origin[1] - size_slow)
                                ) : -1,
                                int(round(subpanel_origin[0])) : int(
                                    round(subpanel_origin[0] - size_fast)
                                ) : -1,
                            ] = source
                            self.active_areas.extend(
                                [
                                    round(subpanel_origin[1] - size_slow),
                                    round(subpanel_origin[0] - size_fast),
                                    round(subpanel_origin[1]),
                                    round(subpanel_origin[0]),
                                ]
                            )
                        else:
                            raise RuntimeError(
                                "Panel angle deviation is too large! Do not use reconst mode!"
                            )

            else:
                size_fast = 1024
                size_slow = 512

                xmin, ymin, xmax, ymax = (
                    0,
                    i * size_fast,
                    size_slow,
                    (i + 1) * size_fast,
                )
                source = data[ymin:ymax, xmin:xmax].transpose()

                if abs(round(self.panel_rotations[i]) + 90) < 1:
                    det[
                        round(origin[1]) : round(origin[1] + size_slow),
                        round(origin[0]) : round(origin[0] + size_fast),
                    ] = source
                elif abs(round(self.panel_rotations[i]) - 90) < 1:
                    det[
                        round(origin[1]) : round(origin[1] - size_slow) : -1,
                        round(origin[0]) : round(origin[0] - size_fast) : -1,
                    ] = source
                else:
                    raise RuntimeError(
                        "Panel angle deviation is too large! Do not use reconst mode!"
                    )

        self.active_areas = [int(aa) for aa in self.active_areas]
        return det

    def get_detector(self, index=None):
        if self._detector_instance is None:
            self._detector_instance = self._detector()

        return self._detector_instance

    def get_static_mask(self):
        # This means when the pixel mask is present, trusted region is ignored.
        # The used provided masks (if any) will be automatically merged.
        # see https://github.com/dials/dials/issues/236
        return self.mask

    def get_beam(self, index=None):
        if index is not None and self.index != index:
            self.set_index(index)
            self._beam_instance = None

        if self._beam_instance is None:
            self._beam_instance = self._beam()

        return self._beam_instance


if __name__ == "__main__":
    print(FormatHDF5SaclaMPCCD.understand(sys.argv[1]))
    FormatHDF5SaclaMPCCD(sys.argv[1])



from __future__ import annotations

import os
import re
from collections import defaultdict
from glob import glob

import natsort


def template_regex(filename):
    """Works out a template from a filename.

    Tries a bunch of templates to work out the most sensible. N.B. assumes
    that the image index will be the last digits found in the file name.

    Arguments:
      filename (str): The filename to template-ize

    Returns:
        Tuple[str or None, int]:
            Where the tuple parts are:

            template:
                The template based on the filename, or None if no template
                could be found for the filename.
            index:
                The template index of the filename, or zero if there
                was no template filename.
    """

    # filename template code stolen from xia2...

    if not hasattr(template_regex, "pattern"):
        # Compile regular expressions once and store compiled version

        # The filename is reversed for pattern evaluation.
        # These are reversed patterns.

        patterns = [
            r"()([0-9]+)(\..*)",
            # filename ends with numbers
            #  img.0815
            r"([a-zA-Z-]+\.)([0-9]+)(\..*)",
            # filename ends with numbers followed by simple extension
            #  img.0815.cbf
            r"(.*?\.)([0-9]+)(_.*)",
            # last number in the filename standing between _ and .
            #  NO2_0100.sequence.18keV
            r"(.*?\.)([0-9]+)(.*)",
            # filename ends with numbers
            #  img0815
            r"()([0-9]+)(.*)",
        ]
        # last number in the filename before a .
        #  NO2.00100.sequence.18keV or image-00001.cbf

        # patterns are evaluated in order, the first to match will be used
        # each pattern needs to describe the entire filename with three braces:
        #  (suffix)(imagenumber)(prefix)
        # if more braces are used in a pattern they MUST start with (?:...) so that
        # they do not interfere with the matching
        template_regex.pattern = [re.compile(pattern) for pattern in patterns]

    rfilename = filename[::-1]

    template = None
    digits = 0
    for j, cp in enumerate(template_regex.pattern):
        match = cp.match(rfilename)
        if not match:
            continue
        groups = match.groups()

        exten = groups[0][::-1]
        digits = groups[1][::-1]
        prefix = groups[2][::-1]

        template = prefix + ("#" * len(digits)) + exten
        break

    # print "File name template:\n    %s\n -> %s (%d)" % (filename, template, int(digits))
    return template, int(digits)


def template_regex_from_list(filenames):
    """Compute the template for a list of filenames

    Arguments:
        filenames (List[str]):
            List of filenames. *Must* be longer than len(1), or a
            TypeError will be raised.

    Returns:
        Tuple[str, List[int]]:
            Where the tuple parts are:

            template: The template describing all entries
            indices:  The template indices present in the collection

    Raises:
        AssertionError:
            If not all filenames resolve to the same regex
        TypeError:
            If none of the filenames resolve to a template, or
            len(filenames) == 1.
    """

    common_prefix = os.path.commonprefix(filenames)
    templates, indices = zip(
        *[template_regex(f[len(common_prefix) :]) for f in filenames]
    )
    template = templates[0]
    assert all(t == template for t in templates[1:])
    return common_prefix + template, indices


def group_files_by_imageset(filenames):
    """Group filenames by supposed imageset.

    Get the template for each file in the list. Then add to a dictionary
    indexed by template containing a list of indices within that template.
    For files that do not match a template, these are added by filename
    instead.

    """

    # Calculate the template for each image. If the template is None
    # (i.e. there are no numbers to identify the filename, add the
    # filename itself.
    template = []
    for f in filenames:
        f = os.fspath(f)
        t = template_regex(f)
        if t[0] is None:
            template.append((f, None))
        else:
            template.append(t)

    # Loop through all the templates and add the new item to a dictionary
    # with a list of files per template.
    matched = defaultdict(list)
    for t in template:
        matched[t[0]].append(t[1])

    return matched


def find_matching_images(image_name):
    """Search in the directory in which this image is for images which share
    the same template: return this list."""

    directory, filename = os.path.split(image_name)
    if directory is None or directory == "":
        directory = "."

    template, digits = template_regex(filename)

    if template:
        len_digits = template.count("#")
        pfx = template.split("#")[0]
        sfx = template.split("#")[-1]
        template_str = pfx + "%%0%dd" % len_digits + sfx

        files_in_directory = os.listdir(directory)

        matching_images = []

        for j in range(0, 10**len_digits):
            if template_str % j in files_in_directory:
                matching_images.append(os.path.join(directory, template_str % j))

    else:
        matching_images = [image_name]

    return matching_images


def replace_template_format_with_hash(match):
    """Replace the format match with hashes"""
    return "#" * len(match.group(0) % 0)


def template_string_to_glob_expr(template):
    """Convert the template to a glob expression."""
    if template.count("#") == 1:
        # https://github.com/cctbx/dxtbx/issues/646
        return template.replace("#", "*")
    return template.replace("#", "[0-9]")


def template_string_number_index(template):
    """Get the index slice ends for `##...#` in the template."""
    return template.find("#"), template.rfind("#") + 1


def locate_files_matching_template_string(template):
    """Return all files matching template."""
    matches = glob(template_string_to_glob_expr(template))
    if template.count("#") != 1:
        return matches
    matches = [os.path.split(p) for p in matches]
    i0, i1 = template_string_number_index(template)
    suffix = template[i1:]
    patt = re.compile("([^0]*)([0-9]+)" + suffix)
    return [os.path.join(*m) for m in matches if patt.match(m[1])]


def template_image_range(template):
    """Return the image range of files with this template."""

    # Find the files matching the template
    filenames = locate_files_matching_template_string(template)
    filenames = natsort.natsorted(filenames)

    # Check that the template matches some files
    if len(filenames) == 0:
        raise ValueError(f"Template {template} doesn't match any files.")

    # Get the first and last indices
    if "#" in template:
        i0, i1 = template_string_number_index(template)
        prefix = template[:i0]
        suffix = template[i1:]
        first = int(filenames[0].replace(prefix, "").replace(suffix, ""))
        last = int(filenames[-1].replace(prefix, "").replace(suffix, ""))
    else:  # template is one file
        first, last = 0, 0

    # Return the image range
    return (first, last)



"""
An implementation of the SMV image reader for Rigaku A200 images.
Inherits from FormatSMVRigaku.
"""

from __future__ import annotations

import sys

from iotbx.detectors.dtrek import DTREKImage
from scitbx import matrix

from dxtbx.format.FormatSMVRigaku import FormatSMVRigaku


class FormatSMVRigakuA200(FormatSMVRigaku):
    """A class for reading SMV format Rigaku A200 images, and correctly
    constructing a model for the experiment from this."""

    @staticmethod
    def understand(image_file):
        """Check to see if this looks like a Rigaku A200 SMV format image,
        i.e. we can make sense of it. Essentially that will be if it contains
        all of the keys we are looking for."""

        size, header = FormatSMVRigaku.get_smv_header(image_file)

        wanted_header_items = [
            "DETECTOR_NUMBER",
            "DETECTOR_NAMES",
            "CRYSTAL_GONIO_NUM_VALUES",
            "CRYSTAL_GONIO_NAMES",
            "CRYSTAL_GONIO_UNITS",
            "CRYSTAL_GONIO_VALUES",
            "ROTATION",
            "ROTATION_AXIS_NAME",
            "ROTATION_VECTOR",
            "SOURCE_VECTORS",
            "SOURCE_WAVELENGTH",
            "SOURCE_POLARZ",
            "DIM",
            "SIZE1",
            "SIZE2",
        ]

        if any(item not in header for item in wanted_header_items):
            return False

        detector_prefix = header["DETECTOR_NAMES"].split()[0].strip()

        more_wanted_header_items = [
            "DETECTOR_DIMENSIONS",
            "DETECTOR_SIZE",
            "DETECTOR_VECTORS",
            "GONIO_NAMES",
            "GONIO_UNITS",
            "GONIO_VALUES",
            "GONIO_VECTORS",
            "SPATIAL_BEAM_POSITION",
            "SPATIAL_DISTORTION_VECTORS",
        ]

        if any(
            f"{detector_prefix}{item}" not in header
            for item in more_wanted_header_items
        ):
            return False

        descriptive_items = ["DETECTOR_IDENTIFICATION", "DETECTOR_DESCRIPTION"]

        for header_item in descriptive_items:
            test = f"{detector_prefix}{header_item}"
            if test in header and (
                "raxis" in header[test].lower() or "a200" in header[test].lower()
            ):
                return True

        return False

    def detectorbase_start(self):
        self.detectorbase = DTREKImage(self._image_file)
        self.detectorbase.readHeader()

    def _goniometer(self):
        """Initialize the structure for the goniometer - this will need to
        correctly compose the axes given in the image header. In this case
        this is made rather straightforward as the image header has the
        calculated rotation axis stored in it. We could work from the
        rest of the header and construct a goniometer model."""

        axis = tuple(map(float, self._header_dictionary["ROTATION_VECTOR"].split()))

        return self._goniometer_factory.known_axis(axis)

    def _detector(self):
        """Return a model for the detector, allowing for two-theta offsets
        and the detector position. This will be rather more complex..."""

        detector_name = self._header_dictionary["DETECTOR_NAMES"].split()[0].strip()

        detector_axes = self.get_detector_axes(detector_name)
        detector_x = matrix.col(tuple(detector_axes[:3]))
        detector_y = matrix.col(tuple(detector_axes[3:]))

        # Now map these to real axes
        distortion = self.get_distortion(detector_name)
        detector_fast = detector_x * distortion[0] + detector_y * distortion[1]
        detector_slow = detector_x * distortion[2] + detector_y * distortion[3]

        beam_pixels = self.get_beam_pixels(detector_name)
        pixel_size = self.get_pixel_size(detector_name)
        image_size = self.get_image_size(detector_name)

        detector_origin = -(
            beam_pixels[0] * pixel_size[0] * detector_fast
            + beam_pixels[1] * pixel_size[1] * detector_slow
        )

        gonio_axes = self.get_gonio_axes(detector_name)
        gonio_values = self.get_gonio_values(detector_name)
        gonio_units = self._header_dictionary["%sGONIO_UNITS" % detector_name].split()
        gonio_num_axes = int(
            self._header_dictionary["%sGONIO_NUM_VALUES" % detector_name]
        )

        rotations = []
        translations = []

        for j, unit in enumerate(gonio_units):
            axis = matrix.col(gonio_axes[3 * j : 3 * (j + 1)])
            if unit == "deg":
                rotations.append(
                    axis.axis_and_angle_as_r3_rotation_matrix(gonio_values[j], deg=True)
                )
                translations.append(matrix.col((0.0, 0.0, 0.0)))
            elif unit == "mm":
                rotations.append(
                    matrix.sqr((1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0))
                )
                translations.append(gonio_values[j] * axis)
            else:
                raise RuntimeError("unknown axis unit %s" % unit)

        rotations.reverse()
        translations.reverse()

        for j in range(gonio_num_axes):
            detector_fast = rotations[j] * detector_fast
            detector_slow = rotations[j] * detector_slow
            detector_origin = rotations[j] * detector_origin
            detector_origin = translations[j] + detector_origin

        max_trusted_value = int(float(self._header_dictionary["SATURATED_VALUE"]))
        min_trusted_value = 0

        return self._detector_factory.complex(
            "CCD",
            detector_origin.elems,
            detector_fast.elems,
            detector_slow.elems,
            pixel_size,
            image_size,
            (min_trusted_value, max_trusted_value),
        )

    def _beam(self):
        """Return a simple model for the beam."""

        beam_direction = self.get_beam_direction()
        p_fraction, p_plane = self.get_beam_polarization()

        wavelength = float(self._header_dictionary["SCAN_WAVELENGTH"])

        return self._beam_factory.complex(
            beam_direction, p_fraction, p_plane, wavelength
        )

    def _scan(self):
        """Return the scan information for this image."""

        rotation = self.get_rotation()
        epoch = 0

        exposure_time = rotation[3]
        osc_start = rotation[0]
        osc_range = rotation[2]

        return self._scan_factory.single_file(
            self._image_file, exposure_time, osc_start, osc_range, epoch
        )


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatSMVRigakuA200.understand(arg))



from __future__ import annotations

from scitbx.array_family import flex

from dxtbx.ext import compress, uncompress


def test_compress_decompress():
    x, y = 10, 10

    data = flex.int(x * y, 1)
    data[10] = 44369
    data[11] = 214
    compressed = compress(data)
    uncompressed = uncompress(compressed, x, y)

    assert list(data) == list(uncompressed)



from __future__ import annotations

import dxtbx


def test_mpccd_nexus_gain(dials_data):
    """
    Tests SACLA MPCCD image from CXI.DB 221
    Includes parameter data_scale_factor, which accounts for a gain of 10
    """

    try:
        h5path = dials_data("image_examples") / "SACLA-MPCCD-run197287-0-nexus.h5"
    except Exception as e:
        print(type(e), str(e))
        raise
    img = dxtbx.load(h5path)

    d = img.get_detector()

    assert d[0].get_gain() == 10



"""Print out the resolution (two-theta) of the corners of the detector"""

from __future__ import annotations

import math
import sys

from scitbx import matrix

import dxtbx


def resolution_corners(frame):
    """Compute the resolution limit corresponding to the corners of the detector
    surface."""

    detector = frame.get_detector()
    beam = frame.get_beam()

    nfast, nslow = map(int, detector.get_image_size())
    dfast, dslow = detector.get_pixel_size()
    F = matrix.col(detector.get_fast_axis())
    S = matrix.col(detector.get_slow_axis())
    origin = matrix.col(detector.get_origin())

    s0 = -1 * matrix.col(beam.get_sample_to_source_direction())

    for ds in 0, 1:
        for df in 0, 1:
            corner = origin + nfast * dfast * F * df + nslow * dslow * S * ds
            theta = 0.5 * corner.angle(s0)
            print("%.3f" % (beam.get_wavelength() / (2 * math.sin(theta))))


if __name__ == "__main__":
    resolution_corners(dxtbx.load(sys.argv[1]))
